# Copyright (c) 2011-2024, Manfred Moitzi
# License: MIT License
from __future__ import annotations
from typing import Iterable, Sequence

from functools import partial
import math
import numpy as np
import numpy.typing as npt

from ezdxf.math import (
    Vec3,
    Vec2,
    UVec,
    Matrix44,
    X_AXIS,
    Y_AXIS,
    arc_angle_span_rad,
)

TOLERANCE = 1e-10
RADIANS_90 = math.pi / 2.0
RADIANS_180 = math.pi
RADIANS_270 = RADIANS_90 * 3.0
RADIANS_360 = 2.0 * math.pi

__all__ = [
    "closest_point",
    "convex_hull_2d",
    "distance_point_line_2d",
    "is_convex_polygon_2d",
    "is_axes_aligned_rectangle_2d",
    "is_point_on_line_2d",
    "is_point_left_of_line",
    "point_to_line_relation",
    "enclosing_angles",
    "sign",
    "area",
    "np_area",
    "circle_radius_3p",
    "TOLERANCE",
    "has_matrix_2d_stretching",
    "decdeg2dms",
    "ellipse_param_span",
]


def sign(f: float) -> float:
    """Return sign of float `f` as -1 or +1, 0 returns +1"""
    return -1.0 if f < 0.0 else +1.0


def decdeg2dms(value: float) -> tuple[float, float, float]:
    """Return decimal degrees as tuple (Degrees, Minutes, Seconds)."""
    mnt, sec = divmod(value * 3600.0, 60.0)
    deg, mnt = divmod(mnt, 60.0)
    return deg, mnt, sec


def ellipse_param_span(start_param: float, end_param: float) -> float:
    """Returns the counter-clockwise params span of an elliptic arc from start-
    to end param.

    Returns the param span in the range [0, 2π], 2π is a full ellipse.
    Full ellipse handling is a special case, because normalization of params
    which describe a full ellipse would return 0 if treated as regular params.
    e.g. (0, 2π) → 2π, (0, -2π) → 2π, (π, -π) → 2π.
    Input params with the same value always return 0 by definition:
    (0, 0) → 0, (-π, -π) → 0, (2π, 2π) → 0.

    Alias to function: :func:`ezdxf.math.arc_angle_span_rad`

    """
    return arc_angle_span_rad(float(start_param), float(end_param))


def closest_point(base: UVec, points: Iterable[UVec]) -> Vec3 | None:
    """Returns the closest point to a give `base` point.

    Args:
        base: base point as :class:`Vec3` compatible object
        points: iterable of points as :class:`Vec3` compatible object

    """
    base = Vec3(base)
    min_dist: float | None = None
    found: Vec3 | None = None
    for point in points:
        p = Vec3(point)
        dist = base.distance(p)
        if (min_dist is None) or (dist < min_dist):
            min_dist = dist
            found = p
    return found


def convex_hull_2d(points: Iterable[UVec]) -> list[Vec2]:
    """Returns the 2D convex hull of given `points`.

    Returns a closed polyline, first vertex is equal to the last vertex.

    Args:
        points: iterable of points, z-axis is ignored

    """

    # Source: https://massivealgorithms.blogspot.com/2019/01/convex-hull-sweep-line.html?m=1
    def cross(o: Vec2, a: Vec2, b: Vec2) -> float:
        return (a - o).det(b - o)

    vertices = Vec2.list(set(points))
    vertices.sort()
    if len(vertices) < 3:
        raise ValueError("Convex hull calculation requires 3 or more unique points.")

    n: int = len(vertices)
    hull: list[Vec2] = [Vec2()] * (2 * n)
    k: int = 0
    i: int
    for i in range(n):
        while k >= 2 and cross(hull[k - 2], hull[k - 1], vertices[i]) <= 0.0:
            k -= 1
        hull[k] = vertices[i]
        k += 1
    t: int = k + 1
    for i in range(n - 2, -1, -1):
        while k >= t and cross(hull[k - 2], hull[k - 1], vertices[i]) <= 0.0:
            k -= 1
        hull[k] = vertices[i]
        k += 1
    return hull[:k]


def enclosing_angles(angle, start_angle, end_angle, ccw=True, abs_tol=TOLERANCE):
    isclose = partial(math.isclose, abs_tol=abs_tol)

    s = start_angle % math.tau
    e = end_angle % math.tau
    a = angle % math.tau
    if isclose(s, e):
        return isclose(s, a)

    if s < e:
        r = s < a < e
    else:
        r = not (e < a < s)
    return r if ccw else not r


def is_point_on_line_2d(
    point: Vec2, start: Vec2, end: Vec2, ray=True, abs_tol=TOLERANCE
) -> bool:
    """Returns ``True`` if `point` is on `line`.

    Args:
        point: 2D point to test as :class:`Vec2`
        start: line definition point as :class:`Vec2`
        end: line definition point as :class:`Vec2`
        ray: if ``True`` point has to be on the infinite ray, if ``False``
            point has to be on the line segment
        abs_tol: tolerance for on the line test

    """
    point_x, point_y = point
    start_x, start_y = start
    end_x, end_y = end
    on_line = (
        math.fabs(
            (end_y - start_y) * point_x
            - (end_x - start_x) * point_y
            + (end_x * start_y - end_y * start_x)
        )
        <= abs_tol
    )
    if not on_line or ray:
        return on_line
    else:
        if start_x > end_x:
            start_x, end_x = end_x, start_x
        if not (start_x - abs_tol <= point_x <= end_x + abs_tol):
            return False
        if start_y > end_y:
            start_y, end_y = end_y, start_y
        if not (start_y - abs_tol <= point_y <= end_y + abs_tol):
            return False
        return True


def point_to_line_relation(
    point: Vec2, start: Vec2, end: Vec2, abs_tol=TOLERANCE
) -> int:
    """Returns ``-1`` if `point` is left `line`, ``+1`` if `point` is right of
    `line` and ``0`` if `point` is on the `line`. The `line` is defined by two
    vertices given as arguments `start` and `end`.

    Args:
        point: 2D point to test as :class:`Vec2`
        start: line definition point as :class:`Vec2`
        end: line definition point as :class:`Vec2`
        abs_tol: tolerance for minimum distance to line

    """
    rel = (end.x - start.x) * (point.y - start.y) - (end.y - start.y) * (
        point.x - start.x
    )
    if abs(rel) <= abs_tol:
        return 0
    elif rel < 0:
        return +1
    else:
        return -1


def is_point_left_of_line(point: Vec2, start: Vec2, end: Vec2, colinear=False) -> bool:
    """Returns ``True`` if `point` is "left of line" defined by `start-` and
    `end` point, a colinear point is also "left of line" if argument `colinear`
    is ``True``.

    Args:
        point: 2D point to test as :class:`Vec2`
        start: line definition point as :class:`Vec2`
        end: line definition point as :class:`Vec2`
        colinear: a colinear point is also "left of line" if ``True``

    """
    rel = point_to_line_relation(point, start, end)
    if colinear:
        return rel < 1
    else:
        return rel < 0


def distance_point_line_2d(point: Vec2, start: Vec2, end: Vec2) -> float:
    """Returns the normal distance from `point` to 2D line defined by `start-`
    and `end` point.
    """
    # wikipedia: https://en.wikipedia.org/wiki/Distance_from_a_point_to_a_line.
    if start.isclose(end):
        raise ZeroDivisionError("Not a line.")
    return math.fabs((start - point).det(end - point)) / (end - start).magnitude


def circle_radius_3p(a: Vec3, b: Vec3, c: Vec3) -> float:
    ba = b - a
    ca = c - a
    cb = c - b
    upper = ba.magnitude * ca.magnitude * cb.magnitude
    lower = ba.cross(ca).magnitude * 2.0
    return upper / lower


def area(vertices: Iterable[UVec]) -> float:
    """Returns the area of a polygon.

    Returns the projected area in the xy-plane for any vertices (z-axis will be ignored).

    """
    # TODO: how to do all this in numpy efficiently?

    vec2s = Vec2.list(vertices)
    if len(vec2s) < 3:
        return 0.0

    # close polygon:
    if not vec2s[0].isclose(vec2s[-1]):
        vec2s.append(vec2s[0])
    return np_area(np.array([(v.x, v.y) for v in vec2s], dtype=np.float64))


def np_area(vertices: npt.NDArray) -> float:
    """Returns the area of a polygon.

    Returns the projected area in the xy-plane, the z-axis will be ignored.
    The polygon has to be closed (first vertex == last vertex) and should have 3 or more
    corner vertices to return a valid result.

    Args:
        vertices: numpy array [:, n], n > 1

    """
    p1x = vertices[:-1, 0]
    p2x = vertices[1:, 0]
    p1y = vertices[:-1, 1]
    p2y = vertices[1:, 1]
    return np.abs(np.sum(p1x * p2y - p1y * p2x)) * 0.5


def has_matrix_2d_stretching(m: Matrix44) -> bool:
    """Returns ``True`` if matrix `m` performs a non-uniform xy-scaling.
    Uniform scaling is not stretching in this context.

    Does not check if the target system is a cartesian coordinate system, use the
    :class:`~ezdxf.math.Matrix44` property :attr:`~ezdxf.math.Matrix44.is_cartesian`
    for that.
    """
    ux = m.transform_direction(X_AXIS)
    uy = m.transform_direction(Y_AXIS)
    return not math.isclose(ux.magnitude_square, uy.magnitude_square)


def is_convex_polygon_2d(polygon: list[Vec2], *, strict=False, epsilon=1e-6) -> bool:
    """Returns ``True`` if the 2D `polygon` is convex.

    This function supports open and closed polygons with clockwise or counter-clockwise
    vertex orientation.

    Coincident vertices will always be skipped and if argument `strict` is ``True``,
    polygons with collinear vertices are not considered as convex.

    This solution works only for simple non-self-intersecting polygons!

    """
    # TODO: Cython implementation
    if len(polygon) < 3:
        return False

    global_sign: int = 0
    current_sign: int = 0
    prev = polygon[-1]
    prev_prev = polygon[-2]
    for vertex in polygon:
        if vertex.isclose(prev):  # skip coincident vertices
            continue

        det = (prev - vertex).det(prev_prev - prev)
        if abs(det) >= epsilon:
            current_sign = -1 if det < 0.0 else +1
            if not global_sign:
                global_sign = current_sign
            # do all determinants have the same sign?
            if global_sign != current_sign:
                return False
        elif strict:  # collinear vertices
            return False

        prev_prev = prev
        prev = vertex
    return bool(global_sign)


def is_axes_aligned_rectangle_2d(points: list[Vec2]) -> bool:
    """Returns ``True`` if the given points represent a rectangle aligned with the
    coordinate system axes.

    The sides of the rectangle must be parallel to the x- and y-axes of the coordinate
    system.  The rectangle can be open or closed (first point == last point) and
    oriented clockwise or counter-clockwise.  Only works with 4 or 5 vertices, rectangles
    that have sides with collinear edges are not considered rectangles.

    .. versionadded:: 1.2.0

    """

    def is_horizontal(a: Vec2, b: Vec2) -> bool:
        return math.isclose(a.y, b.y)

    def is_vertical(a: Vec2, b: Vec2):
        return math.isclose(a.x, b.x)

    count = len(points)
    if points[0].isclose(points[-1]):
        count -= 1
    if count != 4:
        return False
    p0, p1, p2, p3, *_ = points
    if (
        is_horizontal(p0, p1)
        and is_vertical(p1, p2)
        and is_horizontal(p2, p3)
        and is_vertical(p3, p0)
    ):
        return True
    if (
        is_horizontal(p1, p2)
        and is_vertical(p2, p3)
        and is_horizontal(p3, p0)
        and is_vertical(p0, p1)
    ):
        return True
    return False



# Copyright (c) 2019, Manfred Moitzi
# License: MIT License

from ezdxf.math import Vec2, offset_vertices_2d

PRECISION = 1e-3


def test_0_offset():
    vertices = [(1, 2, 3), (5, 2, 3)]
    result = list(offset_vertices_2d(vertices, 0))
    assert result[0] == (1, 2)
    assert result[1] == (5, 2)


def test_2_horiz_vertices_left_offset():
    vertices = [(1, 2), (5, 2)]
    result = list(offset_vertices_2d(vertices, 1))
    assert result[0] == (1, 3)
    assert result[1] == (5, 3)


def test_2_horiz_vertices_right_offset():
    vertices = [(1, 2), (5, 2)]
    result = list(offset_vertices_2d(vertices, -1))
    assert result[0] == (1, 1)
    assert result[1] == (5, 1)


def test_2_vert_vertices_left_offset():
    vertices = [(1, 2), (1, 5)]
    result = list(offset_vertices_2d(vertices, 1))
    assert result[0] == (0, 2)
    assert result[1] == (0, 5)


def test_2_vert_vertices_right_offset():
    vertices = [(1, 2), (1, 5)]
    result = list(offset_vertices_2d(vertices, -1))
    assert result[0] == (2, 2)
    assert result[1] == (2, 5)


def test_3_horiz_collinear_vertices():
    vertices = [(1, 2), (5, 2), (9, 2)]
    result = list(offset_vertices_2d(vertices, 1))
    assert result[0] == (1, 3)
    assert result[1] == (5, 3)
    assert result[2] == (9, 3)


def test_3_vert_collinear_vertices():
    vertices = [(1, 2), (1, 5), (1, 9)]
    result = list(offset_vertices_2d(vertices, 1))
    assert result[0] == (0, 2)
    assert result[1] == (0, 5)
    assert result[2] == (0, 9)


def test_3_vertices():
    vertices = [(0, 0), (300, 150), (450, 50)]
    result = list(offset_vertices_2d(vertices, 10))
    assert result[0].isclose(Vec2((-4.4721, 8.9443)), abs_tol=PRECISION)
    assert result[1].isclose(Vec2((300.7184, 161.5396)), abs_tol=PRECISION)
    assert result[2].isclose(Vec2((455.547, 58.3205)), abs_tol=PRECISION)


def test_closed_square_inside():
    vertices = [(0, 0), (5, 0), (5, 5), (0, 5)]
    result = list(offset_vertices_2d(vertices, 1, closed=True))
    assert result[0] == (1, 1)
    assert result[1] == (4, 1)
    assert result[2] == (4, 4)
    assert result[3] == (1, 4)


def test_closed_triangle_inside():
    vertices = [(0, 0), (5, 0), (2.5, 5)]
    result = list(offset_vertices_2d(vertices, 1, closed=True))
    assert result[0].isclose(Vec2((1.618, 1)), abs_tol=PRECISION)
    assert result[1].isclose(Vec2((3.382, 1)), abs_tol=PRECISION)
    assert result[2].isclose(Vec2((2.5, 2.7639)), abs_tol=PRECISION)


def test_closed_shape_with_collinear_last_segment():
    vertices = [(0, 0), (5, 0), (5, 5), (-5, 5), (-5, 0)]
    result = list(offset_vertices_2d(vertices, 1, closed=True))
    assert len(result) == len(vertices)
    assert result[0] == (0, 1)
    assert result[1] == (4, 1)
    assert result[2] == (4, 4)
    assert result[3] == (-4, 4)
    assert result[4] == (-4, 1)


def test_3_horiz_collinear_vertices_closed():
    vertices = [(1, 2), (5, 2), (9, 2)]
    result = list(offset_vertices_2d(vertices, 1, closed=True))
    assert len(result) == len(vertices) + 2  # get 2 extra vertices
    # but the first vertex, would be expected as last vertex
    assert result[0] == (1, 1)
    assert result[1] == (1, 3)
    assert result[2] == (5, 3)
    # closing segment: (9, 2) -> (1, 2)
    assert result[3] == (9, 3)
    assert result[4] == (9, 1)



#  Copyright (c) 2022, Manfred Moitzi
#  License: MIT License
from __future__ import annotations
from typing import (
    Sequence,
    Iterable,
    Optional,
    Callable,
    Union,
)
import abc
import copy
from dataclasses import dataclass
import json
import random
import time

# example usage:
# examples\addons\optimize\bin_packing_forms.py
# examples\addons\optimize\tsp.py


class DNA(abc.ABC):
    """Abstract DNA class."""

    fitness: Optional[float] = None
    _data: list

    @abc.abstractmethod
    def reset(self, values: Iterable):
        ...

    @property
    @abc.abstractmethod
    def is_valid(self) -> bool:
        ...

    def copy(self):
        return copy.deepcopy(self)

    def _taint(self):
        self.fitness = None

    def __repr__(self):
        return f"{self.__class__.__name__}({str(self._data)})"

    def __eq__(self, other):
        assert isinstance(other, self.__class__)
        return self._data == other._data

    def __len__(self):
        return len(self._data)

    def __getitem__(self, item):
        return self._data.__getitem__(item)

    def __setitem__(self, key, value):
        self._data.__setitem__(key, value)
        self._taint()

    def __iter__(self):
        return iter(self._data)

    @abc.abstractmethod
    def flip_mutate_at(self, index: int) -> None:
        ...


def dna_fitness(dna: DNA) -> float:
    return dna.fitness  # type: ignore


class Mutate(abc.ABC):
    """Abstract mutation."""

    @abc.abstractmethod
    def mutate(self, dna: DNA, rate: float):
        ...


class FlipMutate(Mutate):
    """Flip one bit mutation."""

    def mutate(self, dna: DNA, rate: float):
        for index in range(len(dna)):
            if random.random() < rate:
                dna.flip_mutate_at(index)


class NeighborSwapMutate(Mutate):
    """Swap two neighbors mutation."""

    def mutate(self, dna: DNA, rate: float):
        for index in range(len(dna)):
            if random.random() < rate:
                i2 = index - 1
                tmp = dna[i2]
                dna[i2] = dna[index]
                dna[index] = tmp


class RandomSwapMutate(Mutate):
    """Swap two random places mutation."""

    def mutate(self, dna: DNA, rate: float):
        length = len(dna)
        for index in range(length):
            if random.random() < rate:
                i2 = random.randrange(0, length)
                if i2 == index:
                    i2 -= 1
                tmp = dna[i2]
                dna[i2] = dna[index]
                dna[index] = tmp


class ReverseMutate(Mutate):
    """Reverse some consecutive bits mutation."""

    def __init__(self, bits: int = 3):
        self._bits = int(max(bits, 1))

    def mutate(self, dna: DNA, rate: float):
        length = len(dna)
        if random.random() < rate * (
            length / self._bits
        ):  # applied to all bits at ones
            i1 = random.randrange(length - self._bits)
            i2 = i1 + self._bits
            bits = dna[i1:i2]
            dna[i1:i2] = reversed(bits)


class ScrambleMutate(Mutate):
    """Scramble some consecutive bits mutation."""

    def __init__(self, bits: int = 3):
        self._bits = int(max(bits, 1))

    def mutate(self, dna: DNA, rate: float):
        length = len(dna)
        if random.random() < rate * (
            length / self._bits
        ):  # applied to all bits at ones
            i1 = random.randrange(length - self._bits)
            i2 = i1 + self._bits
            bits = dna[i1:i2]
            random.shuffle(bits)
            dna[i1:i2] = bits


class Mate(abc.ABC):
    """Abstract recombination."""

    @abc.abstractmethod
    def recombine(self, dna1: DNA, dna2: DNA):
        pass


class Mate1pCX(Mate):
    """One point crossover recombination."""

    def recombine(self, dna1: DNA, dna2: DNA):
        length = len(dna1)
        index = random.randrange(0, length)
        recombine_dna_2pcx(dna1, dna2, index, length)


class Mate2pCX(Mate):
    """Two point crossover recombination."""

    def recombine(self, dna1: DNA, dna2: DNA):
        length = len(dna1)
        i1 = random.randrange(0, length)
        i2 = random.randrange(0, length)
        if i1 > i2:
            i1, i2 = i2, i1
        recombine_dna_2pcx(dna1, dna2, i1, i2)


class MateUniformCX(Mate):
    """Uniform recombination."""

    def recombine(self, dna1: DNA, dna2: DNA):
        for index in range(len(dna1)):
            if random.random() > 0.5:
                tmp = dna1[index]
                dna1[index] = dna2[index]
                dna2[index] = tmp


class MateOrderedCX(Mate):
    """Recombination class for ordered DNA like UniqueIntDNA()."""

    def recombine(self, dna1: DNA, dna2: DNA):
        length = len(dna1)
        i1 = random.randrange(0, length)
        i2 = random.randrange(0, length)
        if i1 > i2:
            i1, i2 = i2, i1
        recombine_dna_ocx1(dna1, dna2, i1, i2)


def recombine_dna_2pcx(dna1: DNA, dna2: DNA, i1: int, i2: int) -> None:
    """Two point crossover."""
    part1 = dna1[i1:i2]
    part2 = dna2[i1:i2]
    dna1[i1:i2] = part2
    dna2[i1:i2] = part1


def recombine_dna_ocx1(dna1: DNA, dna2: DNA, i1: int, i2: int) -> None:
    """Ordered crossover."""
    copy1 = dna1.copy()
    replace_dna_ocx1(dna1, dna2, i1, i2)
    replace_dna_ocx1(dna2, copy1, i1, i2)


def replace_dna_ocx1(dna1: DNA, dna2: DNA, i1: int, i2: int) -> None:
    """Replace a part in dna1 by dna2 and preserve order of remaining values in
    dna1.
    """
    old = dna1.copy()
    new = dna2[i1:i2]
    dna1[i1:i2] = new
    index = 0
    new_set = set(new)
    for value in old:
        if value in new_set:
            continue
        if index == i1:
            index = i2
        dna1[index] = value
        index += 1


class FloatDNA(DNA):
    """Arbitrary float numbers in the range [0, 1]."""

    __slots__ = ("_data", "fitness")

    def __init__(self, values: Iterable[float]):
        self._data: list[float] = list(values)
        self._check_valid_data()
        self.fitness: Optional[float] = None

    @classmethod
    def random(cls, length: int) -> FloatDNA:
        return cls((random.random() for _ in range(length)))

    @classmethod
    def n_random(cls, n: int, length: int) -> list[FloatDNA]:
        return [cls.random(length) for _ in range(n)]

    @property
    def is_valid(self) -> bool:
        return all(0.0 <= v <= 1.0 for v in self._data)

    def _check_valid_data(self):
        if not self.is_valid:
            raise ValueError("data value out of range")

    def __str__(self):
        if self.fitness is None:
            fitness = ", fitness=None"
        else:
            fitness = f", fitness={self.fitness:.4f}"
        return f"{str([round(v, 4) for v in self._data])}{fitness}"

    def reset(self, values: Iterable[float]):
        self._data = list(values)
        self._check_valid_data()
        self._taint()

    def flip_mutate_at(self, index: int) -> None:
        self._data[index] = 1.0 - self._data[index]  # flip pick location


class BitDNA(DNA):
    """One bit DNA."""

    __slots__ = ("_data", "fitness")

    def __init__(self, values: Iterable):
        self._data: list[bool] = list(bool(v) for v in values)
        self.fitness: Optional[float] = None

    @property
    def is_valid(self) -> bool:
        return True  # everything can be evaluated to True/False

    @classmethod
    def random(cls, length: int) -> BitDNA:
        return cls(bool(random.randint(0, 1)) for _ in range(length))

    @classmethod
    def n_random(cls, n: int, length: int) -> list[BitDNA]:
        return [cls.random(length) for _ in range(n)]

    def __str__(self):
        if self.fitness is None:
            fitness = ", fitness=None"
        else:
            fitness = f", fitness={self.fitness:.4f}"
        return f"{str([int(v) for v in self._data])}{fitness}"

    def reset(self, values: Iterable) -> None:
        self._data = list(bool(v) for v in values)
        self._taint()

    def flip_mutate_at(self, index: int) -> None:
        self._data[index] = not self._data[index]


class UniqueIntDNA(DNA):
    """Unique integer values in the range from 0 to length-1.
    E.g. UniqueIntDNA(10) = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]

    Requires MateOrderedCX() as recombination class to preserve order and
    validity after DNA recombination.

    Requires mutation by swapping like SwapRandom(), SwapNeighbors(),
    ReversMutate() or ScrambleMutate()
    """

    __slots__ = ("_data", "fitness")

    def __init__(self, values: Union[int, Iterable]):
        self._data: list[int]
        if isinstance(values, int):
            self._data = list(range(values))
        else:
            self._data = [int(v) for v in values]
            if not self.is_valid:
                raise TypeError(self._data)
        self.fitness: Optional[float] = None

    @classmethod
    def random(cls, length: int) -> UniqueIntDNA:
        dna = cls(length)
        random.shuffle(dna._data)
        return dna

    @classmethod
    def n_random(cls, n: int, length: int) -> list[UniqueIntDNA]:
        return [cls.random(length) for _ in range(n)]

    @property
    def is_valid(self) -> bool:
        return len(set(self._data)) == len(self._data)

    def __str__(self):
        if self.fitness is None:
            fitness = ", fitness=None"
        else:
            fitness = f", fitness={self.fitness:.4f}"
        return f"{str([int(v) for v in self._data])}{fitness}"

    def reset(self, values: Iterable) -> None:
        self._data = list(int(v) for v in values)
        self._taint()

    def flip_mutate_at(self, index: int) -> None:
        raise TypeError("flip mutation not supported")


class IntegerDNA(DNA):
    """Integer values in the range from 0 to max_ - 1.
    E.g. IntegerDNA([0, 1, 2, 3, 4, 0, 1, 2, 3, 4], 5)
    """

    __slots__ = ("_data", "fitness")

    def __init__(self, values: Iterable[int], max_: int):
        self._max = int(max_)
        self._data: list[int] = list(values)
        if not self.is_valid:
            raise TypeError(self._data)
        self.fitness: Optional[float] = None

    @classmethod
    def random(cls, length: int, max_: int) -> IntegerDNA:
        imax = int(max_)
        return cls((random.randrange(0, imax) for _ in range(length)), imax)

    @classmethod
    def n_random(cls, n: int, length: int, max_: int) -> list[IntegerDNA]:
        return [cls.random(length, max_) for _ in range(n)]

    @property
    def is_valid(self) -> bool:
        return all(0 <= v < self._max for v in self._data)

    def __repr__(self):
        return f"{self.__class__.__name__}({str(self._data)}, {self._max})"

    def __str__(self):
        if self.fitness is None:
            fitness = ", fitness=None"
        else:
            fitness = f", fitness={self.fitness:.4f}"
        return f"{str([int(v) for v in self._data])}{fitness}"

    def reset(self, values: Iterable) -> None:
        self._data = list(int(v) for v in values)
        self._taint()

    def flip_mutate_at(self, index: int) -> None:
        self._data[index] = self._max - self._data[index] - 1


class Selection(abc.ABC):
    """Abstract selection class."""

    @abc.abstractmethod
    def pick(self, count: int) -> Iterable[DNA]:
        ...

    @abc.abstractmethod
    def reset(self, candidates: Iterable[DNA]):
        ...


class Evaluator(abc.ABC):
    """Abstract evaluation class."""

    @abc.abstractmethod
    def evaluate(self, dna: DNA) -> float:
        ...


class Log:
    @dataclass
    class Entry:
        runtime: float
        fitness: float
        avg_fitness: float

    def __init__(self) -> None:
        self.entries: list[Log.Entry] = []

    def add(self, runtime: float, fitness: float, avg_fitness: float) -> None:
        self.entries.append(Log.Entry(runtime, fitness, avg_fitness))

    def dump(self, filename: str) -> None:
        data = [(e.runtime, e.fitness, e.avg_fitness) for e in self.entries]
        with open(filename, "wt") as fp:
            json.dump(data, fp, indent=4)

    @classmethod
    def load(cls, filename: str) -> Log:
        with open(filename, "rt") as fp:
            data = json.load(fp)
        log = Log()
        for runtime, fitness, avg_fitness in data:
            log.add(runtime, fitness, avg_fitness)
        return log


class HallOfFame:
    def __init__(self, count):
        self.count = count
        self._unique_entries = dict()

    def __iter__(self):
        return (
            self._unique_entries[k] for k in self._sorted_keys()[: self.count]
        )

    def _sorted_keys(self):
        return sorted(self._unique_entries.keys(), reverse=True)

    def add(self, dna: DNA):
        assert dna.fitness is not None
        self._unique_entries[dna.fitness] = dna

    def get(self, count: int) -> list[DNA]:
        entries = self._unique_entries
        keys = self._sorted_keys()
        return [entries[k] for k in keys[: min(count, self.count)]]

    def purge(self):
        if len(self._unique_entries) <= self.count:
            return
        entries = self._unique_entries
        keys = self._sorted_keys()
        self._unique_entries = {k: entries[k] for k in keys[: self.count]}


def threshold_filter(
    candidates: Sequence[DNA], best_fitness: float, threshold: float
) -> Iterable[DNA]:
    if best_fitness <= 0.0:
        minimum = min(candidates, key=dna_fitness)
        min_value = (1.0 - threshold) * minimum.fitness  # type: ignore
    else:
        min_value = best_fitness * threshold
    return (c for c in candidates if c.fitness > min_value)  # type: ignore


class GeneticOptimizer:
    """A genetic algorithm (GA) is a meta-heuristic inspired by the process of
    natural selection. Genetic algorithms are commonly used to generate
    high-quality solutions to optimization and search problems by relying on
    biologically inspired operators such as mutation, crossover and selection.

    Source: https://en.wikipedia.org/wiki/Genetic_algorithm

    This implementation searches always for the maximum fitness, fitness
    comparisons are always done by the "greater than" operator (">").
    The algorithm supports negative values to search for the minimum fitness
    (e.g. Travelling Salesmen Problem: -900 > -1000). Reset the start fitness
    by the method :meth:`reset_fitness` accordingly::

        optimizer.reset_fitness(-1e99)

    """

    def __init__(
        self,
        evaluator: Evaluator,
        max_generations: int,
        max_fitness: float = 1.0,
    ):
        if max_generations < 1:
            raise ValueError("requires max_generations > 0")
        # data:
        self.name = "GeneticOptimizer"
        self.log = Log()
        self.candidates: list[DNA] = []

        # core components:
        self.evaluator: Evaluator = evaluator
        self.selection: Selection = RouletteSelection()
        self.mate: Mate = Mate2pCX()
        self.mutation = FlipMutate()

        # options:
        self.max_generations = int(max_generations)
        self.max_fitness: float = float(max_fitness)
        self.max_runtime: float = 1e99
        self.max_stagnation = 100
        self.crossover_rate = 0.70
        self.mutation_rate = 0.01
        self.elitism: int = 2
        # percentage (0.1 = 10%) of candidates with least fitness to ignore in
        # next generation
        self.threshold: float = 0.0

        # state of last (current) generation:
        self.generation: int = 0
        self.start_time = 0.0
        self.runtime: float = 0.0
        self.best_dna: DNA = BitDNA([])
        self.best_fitness: float = 0.0
        self.stagnation: int = 0  # generations without improvement
        self.hall_of_fame = HallOfFame(10)

    def reset_fitness(self, value: float) -> None:
        self.best_fitness = float(value)

    @property
    def is_executed(self) -> bool:
        return bool(self.generation)

    @property
    def count(self) -> int:
        return len(self.candidates)

    def add_candidates(self, dna: Iterable[DNA]):
        if not self.is_executed:
            self.candidates.extend(dna)
        else:
            raise TypeError("already executed")

    def execute(
        self,
        feedback: Optional[Callable[[GeneticOptimizer], bool]] = None,
        interval: float = 1.0,
    ) -> None:
        if self.is_executed:
            raise TypeError("can only run once")
        if not self.candidates:
            print("no DNA defined!")
        t0 = time.perf_counter()
        self.start_time = t0
        for self.generation in range(1, self.max_generations + 1):
            self.measure_fitness()
            t1 = time.perf_counter()
            self.runtime = t1 - self.start_time
            if (
                self.best_fitness >= self.max_fitness
                or self.runtime >= self.max_runtime
                or self.stagnation >= self.max_stagnation
            ):
                break
            if feedback and t1 - t0 > interval:
                if feedback(self):  # stop if feedback() returns True
                    break
                t0 = t1
            self.next_generation()

    def measure_fitness(self) -> None:
        self.stagnation += 1
        fitness_sum: float = 0.0
        for dna in self.candidates:
            if dna.fitness is not None:
                fitness_sum += dna.fitness
                continue
            fitness = self.evaluator.evaluate(dna)
            dna.fitness = fitness
            fitness_sum += fitness
            self.hall_of_fame.add(dna)
            if fitness > self.best_fitness:
                self.best_fitness = fitness
                self.best_dna = dna
                self.stagnation = 0

        self.hall_of_fame.purge()
        try:
            avg_fitness = fitness_sum / len(self.candidates)
        except ZeroDivisionError:
            avg_fitness = 0.0
        self.log.add(
            time.perf_counter() - self.start_time,
            self.best_fitness,
            avg_fitness,
        )

    def next_generation(self) -> None:
        count = len(self.candidates)
        candidates: list[DNA] = []
        selector = self.selection
        selector.reset(self.filter_threshold(self.candidates))

        if self.elitism > 0:
            candidates.extend(self.hall_of_fame.get(self.elitism))

        while len(candidates) < count:
            dna1, dna2 = selector.pick(2)
            dna1 = dna1.copy()
            dna2 = dna2.copy()
            self.recombine(dna1, dna2)
            self.mutate(dna1, dna2)
            candidates.append(dna1)
            candidates.append(dna2)
        self.candidates = candidates

    def filter_threshold(self, candidates: Sequence[DNA]) -> Iterable[DNA]:
        if self.threshold > 0.0:
            return threshold_filter(
                candidates, self.best_fitness, self.threshold
            )
        else:
            return candidates

    def recombine(self, dna1: DNA, dna2: DNA):
        if random.random() < self.crossover_rate:
            self.mate.recombine(dna1, dna2)

    def mutate(self, dna1: DNA, dna2: DNA):
        self.mutation.mutate(dna1, self.mutation_rate)
        self.mutation.mutate(dna2, self.mutation_rate)


def conv_negative_weights(weights: Iterable[float]) -> Iterable[float]:
    # random.choices does not accept negative values: -100 -> 1/100, -10 -> 1/10
    return (0.0 if w == 0.0 else 1.0 / abs(w) for w in weights)


class RouletteSelection(Selection):
    """Selection by fitness values."""

    def __init__(self, negative_values: bool = False) -> None:
        self._candidates: list[DNA] = []
        self._weights: list[float] = []
        self._negative_values = bool(negative_values)

    def reset(self, candidates: Iterable[DNA]):
        # dna.fitness is not None here!
        self._candidates = list(candidates)
        if self._negative_values:
            self._weights = list(
                conv_negative_weights(dna.fitness for dna in self._candidates)  # type: ignore
            )
        else:
            self._weights = [dna.fitness for dna in self._candidates]  # type: ignore

    def pick(self, count: int) -> Iterable[DNA]:
        return random.choices(self._candidates, self._weights, k=count)


class RankBasedSelection(RouletteSelection):
    """Selection by rank of fitness."""

    def reset(self, candidates: Iterable[DNA]):
        # dna.fitness is not None here!
        self._candidates = list(candidates)
        self._candidates.sort(key=dna_fitness)
        # weight of best_fitness == len(strands)
        # and decreases until 1 for the least fitness
        self._weights = list(range(1, len(self._candidates) + 1))


class TournamentSelection(Selection):
    """Selection by choosing the best of a certain count of candidates."""

    def __init__(self, candidates: int):
        self._candidates: list[DNA] = []
        self.candidates = candidates

    def reset(self, candidates: Iterable[DNA]):
        self._candidates = list(candidates)

    def pick(self, count: int) -> Iterable[DNA]:
        for _ in range(count):
            values = [
                random.choice(self._candidates) for _ in range(self.candidates)
            ]
            values.sort(key=dna_fitness)
            yield values[-1]



#  Copyright (c) 2021-2022, Manfred Moitzi
#  License: MIT License
"""
UserRecord(): store user data in a XRECORD entity.

The group code 302 is used as a structure tag.

All supported data types have a fixed group code:
    - str: 1
    - int: 90 - 32-bit values
    - float: 40 - doubles
    - Vec3, Vec2: 10, 20, 30 - Vec2 is stored as Vec3
    - list, tuple: starts with tag (302, "[") and ends with tag (302, "]")
    - dict: starts with tag (302, "{") and ends with tag (302, "}")

The str type can have a max. length of 2049 characters and cannot contain "\n"
or "\r".

This is an advanced feature for experienced programmers, handle with care!
The attribute UserRecord.data is a simple Python list with read/write access.

The UserRecord can store nested list and dict objects.

BinaryData(): store arbitrary binary data in a XRECORD entity

"""
from __future__ import annotations
from typing import (
    TYPE_CHECKING,
    Optional,
    Iterable,
    Sequence,
    Mapping,
    MutableSequence,
    cast,
)
from ezdxf.lldxf import const
from ezdxf.entities import XRecord
from ezdxf.lldxf.tags import Tags, binary_data_to_dxf_tags
from ezdxf.lldxf.types import dxftag
from ezdxf.math import Vec3, Vec2
from ezdxf.tools import take2
from ezdxf.tools.binarydata import bytes_to_hexstr

if TYPE_CHECKING:
    from ezdxf.document import Drawing

TYPE_GROUP_CODE = 2
STR_GROUP_CODE = 1
INT_GROUP_CODE = 90
FLOAT_GROUP_CODE = 40
VEC3_GROUP_CODE = 10
COLLECTION_GROUP_CODE = 302
START_LIST = "["
END_LIST = "]"
START_DICT = "{"
END_DICT = "}"
DEFAULT_NAME = "UserRecord"

__all__ = ["UserRecord", "BinaryRecord"]


class UserRecord:
    def __init__(
        self,
        xrecord: Optional[XRecord] = None,
        *,
        name: str = DEFAULT_NAME,
        doc: Optional[Drawing] = None,
    ):
        """Setup a :class:`UserRecord` with the given `name`.

        The data is stored in the given `xrecord` object, or in a new created
        :class:`~ezdxf.entities.XRecord` instance if ``None``. If `doc` is not
        ``None`` the new xrecord is added to the OBJECTS section of the DXF
        document.

        Changes of the content has to be committed at the end to be stored in
        the underlying :attr:`xrecord` object.

        Args:
            xrecord (XRecord): underlying :class:`~ezdxf.entities.XRecord` instance,
                if ``None`` a new one will be created
            name (str): name of the user list
            doc (Drawing): DXF document or ``None``

        """
        if xrecord is None:
            if doc is None:
                xrecord = XRecord.new()
            else:
                xrecord = cast(XRecord, doc.objects.new_entity("XRECORD", {}))

        self.xrecord = xrecord
        self.name = str(name)
        self.data: MutableSequence = parse_xrecord(self.xrecord, self.name)

    @property
    def handle(self) -> Optional[str]:
        """DXF handle of the underlying :class:`~ezdxf.entities.XRecord` instance."""
        return self.xrecord.dxf.get("handle")

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.commit()

    def __str__(self):
        """Return str(self)."""
        return str(self.data)

    def commit(self) -> XRecord:
        """Store :attr:`data` in the underlying :class:`~ezdxf.entities.XRecord`
        instance. This call is not required if using the class by the ``with``
        statement.

        Raises:
            DXFValueError: invalid chars ``"\\n"`` or ``"\\r"`` in a string
            DXFTypeError: invalid data type

        """
        self.xrecord.tags = compile_user_record(self.name, self.data)
        return self.xrecord


def parse_xrecord(xrecord: XRecord, name: str) -> list:
    data: list = []
    tags = xrecord.tags
    if tags:
        code, value = tags[0]
        if code != TYPE_GROUP_CODE and value != name:
            raise const.DXFTypeError(
                f"{str(xrecord)} is not an user record of type {name}"
            )
        data.extend(item for item in parse_items(tags[1:]))
    return data


def parse_items(tags: Tags) -> list:
    stack: list = []
    items: list = []
    for tag in tags:
        code, value = tag
        if code == STR_GROUP_CODE:
            items.append(str(value))
        elif code == INT_GROUP_CODE:
            items.append(int(value))
        elif code == FLOAT_GROUP_CODE:
            items.append(float(value))
        elif code == VEC3_GROUP_CODE:
            items.append(Vec3(value))
        elif code == COLLECTION_GROUP_CODE and (
            value == START_LIST or value == START_DICT
        ):
            stack.append(items)
            items = []
        elif code == COLLECTION_GROUP_CODE and (
            value == END_LIST or value == END_DICT
        ):
            try:
                prev_level = stack.pop()
            except IndexError:
                raise const.DXFStructureError(
                    f"invalid nested structure, mismatch of structure tags"
                    f" ({COLLECTION_GROUP_CODE}, ...)"
                )
            if value == END_DICT:
                prev_level.append(dict(take2(items)))
            else:
                prev_level.append(items)
            items = prev_level
        else:
            raise const.DXFValueError(
                f"invalid group code in tag: ({code}, {value})"
            )
    if stack:
        raise const.DXFStructureError(
            f"invalid nested structure, mismatch of structure tags"
            f"({COLLECTION_GROUP_CODE}, ...)"
        )
    return items


def compile_user_record(name: str, data: Iterable) -> Tags:
    tags = Tags()
    tags.append(dxftag(TYPE_GROUP_CODE, name))
    tags.extend(tags_from_list(data))
    return tags


def tags_from_list(items: Iterable) -> Tags:
    tags = Tags()
    for item in items:
        if isinstance(item, str):
            if len(item) > 2049:  # DXF R2000 limit for group codes 0-9
                raise const.DXFValueError(
                    "string too long, max. 2049 characters"
                )
            if "\n" in item or "\r" in item:
                raise const.DXFValueError(
                    "found invalid line break '\\n' or '\\r'"
                )
            tags.append(dxftag(STR_GROUP_CODE, item))
        elif isinstance(item, int):
            tags.append(dxftag(INT_GROUP_CODE, item))
        elif isinstance(item, float):
            tags.append(dxftag(FLOAT_GROUP_CODE, item))
        elif isinstance(item, Vec3):
            tags.append(dxftag(VEC3_GROUP_CODE, item))
        elif isinstance(item, Vec2):
            tags.append(dxftag(VEC3_GROUP_CODE, Vec3(item)))
        elif isinstance(item, Sequence):
            tags.append(dxftag(COLLECTION_GROUP_CODE, START_LIST))
            tags.extend(tags_from_list(item))
            tags.append(dxftag(COLLECTION_GROUP_CODE, END_LIST))
        elif isinstance(item, Mapping):
            tags.append(dxftag(COLLECTION_GROUP_CODE, START_DICT))
            tags.extend(tags_from_list(key_value_list(item)))
            tags.append(dxftag(COLLECTION_GROUP_CODE, END_DICT))
        else:
            raise const.DXFTypeError(f"unsupported type: {type(item)}")
    return tags


def key_value_list(data: Mapping) -> Iterable:
    for k, v in data.items():
        yield k
        yield v


class BinaryRecord:
    def __init__(
        self,
        xrecord: Optional[XRecord] = None,
        *,
        doc: Optional[Drawing] = None,
    ):
        """Setup a :class:`BinaryRecord`.

        The data is stored in the given `xrecord` object, or in a new created
        :class:`~ezdxf.entities.XRecord` instance if ``None``. If `doc` is not
        ``None`` the new xrecord is added to the OBJECTS section of the DXF
        document.

        Changes of the content has to be committed at the end to be stored in
        the underlying :attr:`xrecord` object.

        Args:
            xrecord (XRecord): underlying :class:`~ezdxf.entities.XRecord` instance,
                if ``None`` a new one will be created
            doc (Drawing): DXF document or ``None``

        """
        if xrecord is None:
            if doc is None:
                xrecord = XRecord.new()
            else:
                xrecord = cast(XRecord, doc.objects.new_entity("XRECORD", {}))

        self.xrecord = xrecord
        self.data: bytes = parse_binary_data(self.xrecord.tags)

    @property
    def handle(self) -> Optional[str]:
        """DXF handle of the underlying :class:`~ezdxf.entities.XRecord` instance."""
        return self.xrecord.dxf.get("handle")

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.commit()

    def __str__(self) -> str:
        """Return str(self)."""
        return bytes_to_hexstr(self.data)

    def commit(self) -> XRecord:
        """Store binary :attr:`data` in the underlying :class:`~ezdxf.entities.XRecord`
        instance. This call is not required if using the class by the ``with``
        statement.

        """

        assert isinstance(
            self.data, (bytes, bytearray, memoryview)
        ), "expected binary data"
        self.xrecord.tags = binary_data_to_dxf_tags(
            self.data,
            length_group_code=160,
            value_group_code=310,
            value_size=127,
        )
        return self.xrecord


def parse_binary_data(tags: Tags) -> bytes:
    if tags and tags[0].code == 160:
        return b"".join(t.value for t in tags if t.code == 310)
    else:
        return b""



# Copyright (c) 2019-2024 Manfred Moitzi
# License: MIT License
from __future__ import annotations
from typing import TYPE_CHECKING, Iterable, Optional, Iterator
from typing_extensions import Self
import copy
from ezdxf.math import Vec3, Matrix44
from ezdxf.lldxf.tags import Tags, group_tags
from ezdxf.lldxf.attributes import (
    DXFAttr,
    DXFAttributes,
    DefSubclass,
    XType,
    group_code_mapping,
)
from ezdxf.lldxf import const
from ezdxf.entities import factory
from .dxfentity import base_class, SubclassProcessor, DXFEntity, DXFTagStorage
from .dxfgfx import DXFGraphic, acdb_entity
from .dxfobj import DXFObject
from .objectcollection import ObjectCollection
from .copy import default_copy

if TYPE_CHECKING:
    from ezdxf.entities import DXFNamespace
    from ezdxf.lldxf.tagwriter import AbstractTagWriter
    from ezdxf.document import Drawing

__all__ = [
    "AcadTable",
    "AcadTableBlockContent",
    "acad_table_to_block",
    "read_acad_table_content",
]


acdb_block_reference = DefSubclass(
    "AcDbBlockReference",
    {
        # Block name: an anonymous block begins with a *T value
        "geometry": DXFAttr(2),
        # Insertion point:
        "insert": DXFAttr(10, xtype=XType.point3d, default=Vec3(0, 0, 0)),
    },
)
acdb_block_reference_group_codes = group_code_mapping(acdb_block_reference)

acdb_table = DefSubclass(
    "AcDbTable",
    {
        # Table data version number: 0 = 2010
        "version": DXFAttr(280),
        # Hard of the TABLESTYLE object:
        "table_style_id": DXFAttr(342),
        # Handle of the associated anonymous BLOCK containing the graphical
        # representation:
        "block_record_handle": DXFAttr(343),
        # Horizontal direction vector:
        "horizontal_direction": DXFAttr(11),
        # Flag for table value (unsigned integer):
        "table_value": DXFAttr(90),
        # Number of rows:
        "n_rows": DXFAttr(91),
        # Number of columns:
        "n_cols": DXFAttr(92),
        # Flag for an override:
        "override_flag": DXFAttr(93),
        # Flag for an override of border color:
        "border_color_override_flag": DXFAttr(94),
        # Flag for an override of border lineweight:
        "border_lineweight_override_flag": DXFAttr(95),
        # Flag for an override of border visibility:
        "border_visibility_override_flag": DXFAttr(96),
        # 141: Row height; this value is repeated, 1 value per row
        # 142: Column height; this value is repeated, 1 value per column
        # for every cell:
        #      171: Cell type; this value is repeated, 1 value per cell:
        #           1 = text type
        #           2 = block type
        #      172: Cell flag value; this value is repeated, 1 value per cell
        #      173: Cell merged value; this value is repeated, 1 value per cell
        #      174: Boolean flag indicating if the autofit option is set for the
        #           cell; this value is repeated, 1 value per cell
        #      175: Cell border width (applicable only for merged cells); this
        #           value is repeated, 1 value per cell
        #      176: Cell border height (applicable for merged cells); this value
        #           is repeated, 1 value per cell
        #       91: Cell override flag; this value is repeated, 1 value per cell
        #           (from AutoCAD 2007)
        #      178: Flag value for a virtual edge
        #      145: Rotation value (real; applicable for a block-type cell and
        #           a text-type cell)
        #      344: Hard pointer ID of the FIELD object. This applies only to a
        #           text-type cell. If the text in the cell contains one or more
        #           fields, only the ID of the FIELD object is saved.
        #           The text string (group codes 1 and 3) is ignored
        #        1: Text string in a cell. If the string is shorter than 250
        #           characters, all characters appear in code 1.
        #           If the string is longer than 250 characters, it is divided
        #           into chunks of 250 characters.
        #           The chunks are contained in one or more code 2 codes.
        #           If code 2 codes are used, the last group is a code 1 and is
        #           shorter than 250 characters.
        #           This value applies only to text-type cells and is repeated,
        #           1 value per cell
        #        2: Text string in a cell, in 250-character chunks; optional.
        #           This value applies only to text-type cells and is repeated,
        #           1 value per cell
        #      340: Hard-pointer ID of the block table record.
        #           This value applies only to block-type cells and is repeated,
        #           1 value per cell
        #      144: Block scale (real). This value applies only to block-type
        #           cells and is repeated, 1 value per cell
        #      176: Number of attribute definitions in the block table record
        #           (applicable only to a block-type cell)
        #      for every ATTDEF:
        #           331: Soft pointer ID of the attribute definition in the
        #                block table record, referenced by group code 179
        #                (applicable only for a block-type cell). This value is
        #                repeated once per attribute definition
        #           300: Text string value for an attribute definition, repeated
        #                once per attribute definition and applicable only for
        #                a block-type cell
        #        7: Text style name (string); override applied at the cell level
        #      140: Text height value; override applied at the cell level
        #      170: Cell alignment value; override applied at the cell level
        #       64: Value for the color of cell content; override applied at the
        #           cell level
        #       63: Value for the background (fill) color of cell content;
        #           override applied at the cell level
        #       69: True color value for the top border of the cell;
        #           override applied at the cell level
        #       65: True color value for the right border of the cell;
        #           override applied at the cell level
        #       66: True color value for the bottom border of the cell;
        #           override applied at the cell level
        #       68: True color value for the left border of the cell;
        #           override applied at the cell level
        #      279: Lineweight for the top border of the cell;
        #           override applied at the cell level
        #      275: Lineweight for the right border of the cell;
        #           override applied at the cell level
        #      276: Lineweight for the bottom border of the cell;
        #           override applied at the cell level
        #      278: Lineweight for the left border of the cell;
        #           override applied at the cell level
        #      283: Boolean flag for whether the fill color is on;
        #           override applied at the cell level
        #      289: Boolean flag for the visibility of the top border of the cell;
        #           override applied at the cell level
        #      285: Boolean flag for the visibility of the right border of the cell;
        #           override applied at the cell level
        #      286: Boolean flag for the visibility of the bottom border of the cell;
        #           override applied at the cell level
        #      288: Boolean flag for the visibility of the left border of the cell;
        #           override applied at the cell level
        #       70: Flow direction;
        #           override applied at the table entity level
        #       40: Horizontal cell margin;
        #           override applied at the table entity level
        #       41: Vertical cell margin;
        #           override applied at the table entity level
        #      280: Flag for whether the title is suppressed;
        #           override applied at the table entity level
        #      281: Flag for whether the header row is suppressed;
        #           override applied at the table entity level
        #        7: Text style name (string);
        #           override applied at the table entity level.
        #           There may be one entry for each cell type
        #      140: Text height (real);
        #           override applied at the table entity level.
        #           There may be one entry for each cell type
        #      170: Cell alignment (integer);
        #           override applied at the table entity level.
        #           There may be one entry for each cell type
        #       63: Color value for cell background or for the vertical, left
        #           border of the table; override applied at the table entity
        #           level. There may be one entry for each cell type
        #       64: Color value for cell content or for the horizontal, top
        #           border of the table; override applied at the table entity
        #           level. There may be one entry for each cell type
        #       65: Color value for the horizontal, inside border lines;
        #           override applied at the table entity level
        #       66: Color value for the horizontal, bottom border lines;
        #           override applied at the table entity level
        #       68: Color value for the vertical, inside border lines;
        #           override applied at the table entity level
        #       69: Color value for the vertical, right border lines;
        #           override applied at the table entity level
        #      283: Flag for whether background color is enabled (default = 0);
        #           override applied at the table entity level.
        #           There may be one entry for each cell type: 0/1 = Disabled/Enabled
        #      274-279: Lineweight for each border type of the cell (default = kLnWtByBlock);
        #               override applied at the table entity level.
        #               There may be one group for each cell type
        #      284-289: Flag for visibility of each border type of the cell (default = 1);
        #               override applied at the table entity level.
        #               There may be one group for each cell type: 0/1 = Invisible/Visible
        #  97: Standard/title/header row data type
        #  98: Standard/title/header row unit type
        #   4: Standard/title/header row format string
        #
        # AutoCAD 2007 and before:
        # 177: Cell override flag value (before AutoCAD 2007)
        #  92: Extended cell flags (from AutoCAD 2007), COLLISION: group code
        #      also used by n_cols
        # 301: Text string in a cell. If the string is shorter than 250
        #      characters, all characters appear in code 302.
        #      If the string is longer than 250 characters, it is divided into
        #      chunks of 250 characters.
        #      The chunks are contained in one or more code 303 codes.
        #      If code 393 codes are used, the last group is a code 1 and is
        #      shorter than 250 characters.
        #      --- WRONG: The text is divided into chunks of group code 2 and the last
        #          chuck has group code 1.
        #      This value applies only to text-type cells and is repeated,
        #      1 value per cell (from AutoCAD 2007)
        # 302: Text string in a cell, in 250-character chunks; optional.
        #      This value applies only to text-type cells and is repeated,
        #      302 value per cell (from AutoCAD 2007)
        #      --- WRONG: 302 contains all the text as a long string, tested with more
        #          than 66000 characters
        # BricsCAD writes long text in cells with both methods: 302 & (2, 2, 2, ..., 1)
        #
        # REMARK from Autodesk:
        # Group code 178 is a flag value for a virtual edge. A virtual edge is
        # used when a grid line is shared by two cells.
        # For example, if a table contains one row and two columns and it
        # contains cell A and cell B, the central grid line
        # contains the right edge of cell A and the left edge of cell B.
        # One edge is real, and the other edge is virtual.
        # The virtual edge points to the real edge; both edges have the same
        # set of properties, including color, lineweight, and visibility.
    },
)
acdb_table_group_codes = group_code_mapping(acdb_table)


# todo: implement ACAD_TABLE
class AcadTable(DXFGraphic):
    """DXF ACAD_TABLE entity"""

    DXFTYPE = "ACAD_TABLE"
    DXFATTRIBS = DXFAttributes(
        base_class, acdb_entity, acdb_block_reference, acdb_table
    )
    MIN_DXF_VERSION_FOR_EXPORT = const.DXF2007

    def __init__(self):
        super().__init__()
        self.data = None

    def copy_data(self, entity: Self, copy_strategy=default_copy) -> None:
        """Copy data."""
        assert isinstance(entity, AcadTable)
        entity.data = copy.deepcopy(self.data)

    def load_dxf_attribs(
        self, processor: Optional[SubclassProcessor] = None
    ) -> DXFNamespace:
        dxf = super().load_dxf_attribs(processor)
        if processor:
            processor.fast_load_dxfattribs(
                dxf, acdb_block_reference_group_codes, subclass=2
            )
            tags = processor.fast_load_dxfattribs(
                dxf, acdb_table_group_codes, subclass=3, log=False
            )
            self.load_table(tags)
        return dxf

    def load_table(self, tags: Tags):
        pass

    def export_entity(self, tagwriter: AbstractTagWriter) -> None:
        """Export entity specific data as DXF tags."""
        super().export_entity(tagwriter)
        tagwriter.write_tag2(const.SUBCLASS_MARKER, acdb_block_reference.name)
        self.dxf.export_dxf_attribs(tagwriter, ["geometry", "insert"])
        tagwriter.write_tag2(const.SUBCLASS_MARKER, acdb_table.name)
        self.export_table(tagwriter)

    def export_table(self, tagwriter: AbstractTagWriter):
        pass

    def __referenced_blocks__(self) -> Iterable[str]:
        """Support for "ReferencedBlocks" protocol."""
        if self.doc:
            block_record_handle = self.dxf.get("block_record_handle", None)
            if block_record_handle:
                return (block_record_handle,)
        return tuple()


acdb_table_style = DefSubclass(
    "AcDbTableStyle",
    {
        # Table style version: 0 = 2010
        "version": DXFAttr(280),
        # Table style description (string; 255 characters maximum):
        "name": DXFAttr(3),
        # FlowDirection (integer):
        # 0 = Down
        # 1 = Up
        "flow_direction": DXFAttr(7),
        # Flags (bit-coded)
        "flags": DXFAttr(7),
        # Horizontal cell margin (real; default = 0.06)
        "horizontal_cell_margin": DXFAttr(40),
        # Vertical cell margin (real; default = 0.06)
        "vertical_cell_margin": DXFAttr(41),
        # Flag for whether the title is suppressed:
        # 0/1 = not suppressed/suppressed
        "suppress_title": DXFAttr(280),
        # Flag for whether the column heading is suppressed:
        # 0/1 = not suppressed/suppressed
        "suppress_column_header": DXFAttr(281),
        # The following group codes are repeated for every cell in the table
        #   7: Text style name (string; default = STANDARD)
        # 140: Text height (real)
        # 170: Cell alignment (integer)
        #  62: Text color (integer; default = BYBLOCK)
        #  63: Cell fill color (integer; default = 7)
        # 283: Flag for whether background color is enabled (default = 0):
        #      0/1 = disabled/enabled
        #  90: Cell data type
        #  91: Cell unit type
        # 274-279: Lineweight associated with each border type of the cell
        #          (default = kLnWtByBlock)
        # 284-289: Flag for visibility associated with each border type of the cell
        #          (default = 1): 0/1 = Invisible/Visible
        # 64-69: Color value associated with each border type of the cell
        #        (default = BYBLOCK)
    },
)


# todo: implement TABLESTYLE
class TableStyle(DXFObject):
    """DXF TABLESTYLE entity

    Every ACAD_TABLE has its own table style.

    Requires DXF version AC1021/R2007
    """

    DXFTYPE = "TABLESTYLE"
    DXFATTRIBS = DXFAttributes(base_class, acdb_table_style)
    MIN_DXF_VERSION_FOR_EXPORT = const.DXF2007


class TableStyleManager(ObjectCollection[TableStyle]):
    def __init__(self, doc: Drawing):
        super().__init__(doc, dict_name="ACAD_TABLESTYLE", object_type="TABLESTYLE")


@factory.register_entity
class AcadTableBlockContent(DXFTagStorage):
    DXFTYPE = "ACAD_TABLE"
    DXFATTRIBS = DXFAttributes(
        base_class, acdb_entity, acdb_block_reference, acdb_table
    )

    def load_dxf_attribs(
        self, processor: Optional[SubclassProcessor] = None
    ) -> DXFNamespace:
        dxf = super().load_dxf_attribs(processor)
        if processor:
            processor.fast_load_dxfattribs(
                dxf, acdb_block_reference_group_codes, subclass=2
            )
            processor.fast_load_dxfattribs(
                dxf, acdb_table_group_codes, subclass=3, log=False
            )
        return dxf

    def proxy_graphic_content(self) -> Iterable[DXFGraphic]:
        return super().__virtual_entities__()

    def _block_content(self) -> Iterator[DXFGraphic]:
        block_name: str = self.get_block_name()
        return self.doc.blocks.get(block_name, [])  # type: ignore

    def get_block_name(self) -> str:
        return self.dxf.get("geometry", "")

    def get_insert_location(self) -> Vec3:
        return self.dxf.get("insert", Vec3())

    def __virtual_entities__(self) -> Iterator[DXFGraphic]:
        """Implements the SupportsVirtualEntities protocol."""
        insert: Vec3 = Vec3(self.get_insert_location())
        m: Optional[Matrix44] = None
        if insert:
            # TODO: OCS transformation (extrusion) is ignored yet
            m = Matrix44.translate(insert.x, insert.y, insert.z)

        for entity in self._block_content():
            try:
                clone = entity.copy()
            except const.DXFTypeError:
                continue
            if m is not None:
                try:
                    clone.transform(m)
                except:  # skip entity at any transformation issue
                    continue
            yield clone


def acad_table_to_block(table: DXFEntity) -> None:
    """Converts the given ACAD_TABLE entity to a block references (INSERT entity).

    The original ACAD_TABLE entity will be destroyed.

    .. versionadded:: 1.1

    """
    if not isinstance(table, AcadTableBlockContent):
        return
    doc = table.doc
    owner = table.dxf.owner
    block_name = table.get_block_name()
    if doc is None or block_name == "" or owner is None:
        return
    try:
        layout = doc.layouts.get_layout_by_key(owner)
    except const.DXFKeyError:
        return
    # replace ACAD_TABLE entity by INSERT entity
    layout.add_blockref(
        block_name,
        insert=table.get_insert_location(),
        dxfattribs={"layer": table.dxf.get("layer", "0")},
    )
    layout.delete_entity(table)  # type: ignore


def read_acad_table_content(table: DXFTagStorage) -> list[list[str]]:
    """Returns the content of an ACAD_TABLE entity as list of table rows.

    If the count of table rows or table columns is missing the complete content is
    stored in the first row.
    """
    if table.dxftype() != "ACAD_TABLE":
        raise const.DXFTypeError(f"Expected ACAD_TABLE entity, got {str(table)}")
    acdb_table = table.xtags.get_subclass("AcDbTable")

    nrows = acdb_table.get_first_value(91, 0)
    ncols = acdb_table.get_first_value(92, 0)
    split_code = 171  # DXF R2004
    if acdb_table.has_tag(302):
        split_code = 301  # DXF R2007 and later
    values = _load_table_values(acdb_table, split_code)
    if nrows * ncols == 0:
        return [values]
    content: list[list[str]] = []
    for index in range(nrows):
        start = index * ncols
        content.append(values[start : start + ncols])
    return content


def _load_table_values(tags: Tags, split_code: int) -> list[str]:
    values: list[str] = []
    for group in group_tags(tags, splitcode=split_code):
        g_tags = Tags(group)
        if g_tags.has_tag(302):  # DXF R2007 and later
            # contains all text as one long string, with more than 66000 chars tested
            values.append(g_tags.get_first_value(302))
        else:  
            # DXF R2004
            # Text is divided into chunks (2, 2, 2, ..., 1) or (3, 3, 3, ..., 1)
            # DXF reference says group code 2, BricsCAD writes group code 3
            s = "".join(tag.value for tag in g_tags if 1 <= tag.code <= 3)
            values.append(s)
    return values



#  Copyright (c) 2023, Manfred Moitzi
#  License: MIT License
from __future__ import annotations
from typing import Iterable, Sequence, no_type_check

import copy
from xml.etree import ElementTree as ET
import numpy as np

from ezdxf.math import Vec2, BoundingBox2d, Matrix44
from ezdxf.path import Command


from .type_hints import Color
from .backend import BackendInterface, BkPath2d, BkPoints2d, ImageData
from .config import Configuration, LineweightPolicy
from .properties import BackendProperties
from . import layout, recorder

__all__ = ["SVGBackend"]


class SVGBackend(recorder.Recorder):
    """This is a native SVG rendering backend and does not require any external packages
    to render SVG images other than the core dependencies.  This backend support content
    cropping at page margins.
    """

    def __init__(self) -> None:
        super().__init__()
        self._init_flip_y = True
        self.transformation_matrix: Matrix44 | None = None

    def get_xml_root_element(
        self,
        page: layout.Page,
        *,
        settings: layout.Settings = layout.Settings(),
        render_box: BoundingBox2d | None = None,
    ) -> ET.Element:
        top_origin = True
        settings = copy.copy(settings)

        # This player changes the original recordings!
        player = self.player()
        if render_box is None:
            render_box = player.bbox()

        # the page origin (0, 0) is in the top-left corner.
        output_layout = layout.Layout(render_box, flip_y=self._init_flip_y)
        page = output_layout.get_final_page(page, settings)
        if page.width == 0 or page.height == 0:
            return ET.Element("svg")  # empty page

        self.transformation_matrix = output_layout.get_placement_matrix(
            page, settings=settings, top_origin=top_origin
        )
        # transform content to the output coordinates space:
        player.transform(self.transformation_matrix)
        if settings.crop_at_margins:
            p1, p2 = page.get_margin_rect(top_origin=top_origin)  # in mm
            # scale factor to map page coordinates to output space coordinates:
            output_scale = settings.page_output_scale_factor(page)
            max_sagitta = 0.1 * output_scale  # curve approximation 0.1 mm
            # crop content inplace by the margin rect:
            player.crop_rect(p1 * output_scale, p2 * output_scale, max_sagitta)

        self._init_flip_y = False
        backend = self.make_backend(page, settings)
        player.replay(backend)
        return backend.get_xml_root_element()

    def get_string(
        self,
        page: layout.Page,
        *,
        settings: layout.Settings = layout.Settings(),
        render_box: BoundingBox2d | None = None,
        xml_declaration=True,
    ) -> str:
        """Returns the XML data as unicode string.

        Args:
            page: page definition, see :class:`~ezdxf.addons.drawing.layout.Page`
            settings: layout settings, see :class:`~ezdxf.addons.drawing.layout.Settings`
            render_box: set explicit region to render, default is content bounding box
            xml_declaration: inserts the "<?xml version='1.0' encoding='utf-8'?>" string
                in front of the <svg> element

        """
        xml = self.get_xml_root_element(page, settings=settings, render_box=render_box)
        return ET.tostring(xml, encoding="unicode", xml_declaration=xml_declaration)

    @staticmethod
    def make_backend(page: layout.Page, settings: layout.Settings) -> SVGRenderBackend:
        """Override this method to use a customized render backend."""
        return SVGRenderBackend(page, settings)


def make_view_box(page: layout.Page, output_coordinate_space: float) -> tuple[int, int]:
    size = round(output_coordinate_space)
    if page.width > page.height:
        return size, round(size * (page.height / page.width))
    return round(size * (page.width / page.height)), size


def scale_page_to_view_box(page: layout.Page, output_coordinate_space: float) -> float:
    # The viewBox coordinates are integer values in the range of [0, output_coordinate_space]
    return min(
        output_coordinate_space / page.width,
        output_coordinate_space / page.height,
    )


class Styles:
    def __init__(self, xml: ET.Element) -> None:
        self._xml = xml
        self._class_names: dict[int, str] = dict()
        self._counter = 1

    def get_class(
        self,
        *,
        stroke: Color = "none",
        stroke_width: int | str = "none",
        stroke_opacity: float = 1.0,
        fill: Color = "none",
        fill_opacity: float = 1.0,
    ) -> str:
        style = (
            f"{{stroke: {stroke}; "
            f"stroke-width: {stroke_width}; "
            f"stroke-opacity: {stroke_opacity:.3f}; "
            f"fill: {fill}; "
            f"fill-opacity: {fill_opacity:.3f};}}"
        )
        key = hash(style)
        try:
            return self._class_names[key]
        except KeyError:
            pass
        name = f"C{self._counter:X}"
        self._counter += 1
        self._add_class(name, style)
        self._class_names[key] = name
        return name

    def _add_class(self, name, style_str: str) -> None:
        style = ET.Element("style")
        style.text = f".{name} {style_str}"
        self._xml.append(style)


CMD_M_ABS = "M {0.x:.0f} {0.y:.0f}"
CMD_M_REL = "m {0.x:.0f} {0.y:.0f}"
CMD_L_ABS = "L {0.x:.0f} {0.y:.0f}"
CMD_L_REL = "l {0.x:.0f} {0.y:.0f}"
CMD_C3_ABS = "Q {0.x:.0f} {0.y:.0f} {1.x:.0f} {1.y:.0f}"
CMD_C3_REL = "q {0.x:.0f} {0.y:.0f} {1.x:.0f} {1.y:.0f}"
CMD_C4_ABS = "C {0.x:.0f} {0.y:.0f} {1.x:.0f} {1.y:.0f} {2.x:.0f} {2.y:.0f}"
CMD_C4_REL = "c {0.x:.0f} {0.y:.0f} {1.x:.0f} {1.y:.0f} {2.x:.0f} {2.y:.0f}"
CMD_CONT = "{0.x:.0f} {0.y:.0f}"


class SVGRenderBackend(BackendInterface):
    """Creates the SVG output.

    This backend requires some preliminary work, record the frontend output via the
    Recorder backend to accomplish the following requirements:

    - Scale the content in y-axis by -1 to invert the y-axis (SVG).
    - Move content in the first quadrant of the coordinate system.
    - The viewBox is defined by the lower left corner in the origin (0, 0) and
      the upper right corner at (view_box_width, view_box_height)
    - The output coordinates are integer values, scale the content appropriately.
    - Replay the recorded output on this backend.

    """

    def __init__(self, page: layout.Page, settings: layout.Settings) -> None:
        self.settings = settings
        self._stroke_width_cache: dict[float, int] = dict()
        view_box_width, view_box_height = make_view_box(
            page, settings.output_coordinate_space
        )
        # StrokeWidthPolicy.absolute:
        # stroke-width in mm as resolved by the frontend
        self.stroke_width_scale: float = view_box_width / page.width_in_mm
        self.min_lineweight = 0.05  # in mm, set by configure()
        self.lineweight_scaling = 1.0  # set by configure()
        self.lineweight_policy = LineweightPolicy.ABSOLUTE  # set by configure()
        # fixed lineweight for all strokes in ABSOLUTE mode:
        # set Configuration.min_lineweight to the desired lineweight in 1/300 inch!
        # set Configuration.lineweight_scaling to 0

        # LineweightPolicy.RELATIVE:
        # max_stroke_width is determined as a certain percentage of settings.output_coordinate_space
        self.max_stroke_width: int = int(
            settings.output_coordinate_space * settings.max_stroke_width
        )
        # min_stroke_width is determined as a certain percentage of max_stroke_width
        self.min_stroke_width: int = int(
            self.max_stroke_width * settings.min_stroke_width
        )
        # LineweightPolicy.RELATIVE_FIXED:
        # all strokes have a fixed stroke-width as a certain percentage of max_stroke_width
        self.fixed_stroke_width: int = int(
            self.max_stroke_width * settings.fixed_stroke_width
        )
        self.root = ET.Element(
            "svg",
            xmlns="http://www.w3.org/2000/svg",
            width=f"{page.width_in_mm:g}mm",
            height=f"{page.height_in_mm:g}mm",
            viewBox=f"0 0 {view_box_width} {view_box_height}",
        )
        self.styles = Styles(ET.SubElement(self.root, "defs"))
        self.background = ET.SubElement(
            self.root,
            "rect",
            fill="white",
            x="0",
            y="0",
            width=str(view_box_width),
            height=str(view_box_height),
        )
        self.entities = ET.SubElement(self.root, "g")
        self.entities.set("stroke-linecap", "round")
        self.entities.set("stroke-linejoin", "round")
        self.entities.set("fill-rule", "evenodd")

    def get_xml_root_element(self) -> ET.Element:
        return self.root

    def add_strokes(self, d: str, properties: BackendProperties):
        if not d:
            return
        element = ET.SubElement(self.entities, "path", d=d)
        stroke_width = self.resolve_stroke_width(properties.lineweight)
        stroke_color, stroke_opacity = self.resolve_color(properties.color)
        cls = self.styles.get_class(
            stroke=stroke_color,
            stroke_width=stroke_width,
            stroke_opacity=stroke_opacity,
        )
        element.set("class", cls)

    def add_filling(self, d: str, properties: BackendProperties):
        if not d:
            return
        element = ET.SubElement(self.entities, "path", d=d)
        fill_color, fill_opacity = self.resolve_color(properties.color)
        cls = self.styles.get_class(fill=fill_color, fill_opacity=fill_opacity)
        element.set("class", cls)

    def resolve_color(self, color: Color) -> tuple[Color, float]:
        return color[:7], alpha_to_opacity(color[7:9])

    def resolve_stroke_width(self, width: float) -> int:
        try:
            return self._stroke_width_cache[width]
        except KeyError:
            pass
            stroke_width = self.fixed_stroke_width
        policy = self.lineweight_policy
        if policy == LineweightPolicy.ABSOLUTE:
            if self.lineweight_scaling:
                width = max(self.min_lineweight, width) * self.lineweight_scaling
            else:
                width = self.min_lineweight
            stroke_width = round(width * self.stroke_width_scale)
        elif policy == LineweightPolicy.RELATIVE:
            stroke_width = map_lineweight_to_stroke_width(
                width, self.min_stroke_width, self.max_stroke_width
            )
        self._stroke_width_cache[width] = stroke_width
        return stroke_width

    def set_background(self, color: Color) -> None:
        color_str = color[:7]
        opacity = alpha_to_opacity(color[7:9])
        self.background.set("fill", color_str)
        self.background.set("fill-opacity", str(opacity))

    def draw_point(self, pos: Vec2, properties: BackendProperties) -> None:
        self.add_strokes(self.make_polyline_str([pos, pos]), properties)

    def draw_line(self, start: Vec2, end: Vec2, properties: BackendProperties) -> None:
        self.add_strokes(self.make_polyline_str([start, end]), properties)

    def draw_solid_lines(
        self, lines: Iterable[tuple[Vec2, Vec2]], properties: BackendProperties
    ) -> None:
        lines = list(lines)
        if len(lines) == 0:
            return
        self.add_strokes(self.make_multi_line_str(lines), properties)

    def draw_path(self, path: BkPath2d, properties: BackendProperties) -> None:
        self.add_strokes(self.make_path_str(path), properties)

    def draw_filled_paths(
        self, paths: Iterable[BkPath2d], properties: BackendProperties
    ) -> None:
        d = []
        for path in paths:
            if len(path):
                d.append(self.make_path_str(path, close=True))
        self.add_filling(" ".join(d), properties)

    def draw_filled_polygon(
        self, points: BkPoints2d, properties: BackendProperties
    ) -> None:
        self.add_filling(
            self.make_polyline_str(points.vertices(), close=True), properties
        )

    def draw_image(self, image_data: ImageData, properties: BackendProperties) -> None:
        pass  # TODO: not implemented

    @staticmethod
    def make_polyline_str(points: Sequence[Vec2], close=False) -> str:
        if len(points) < 2:
            return ""
        current = points[0]
        # first move is absolute, consecutive lines are relative:
        d: list[str] = [CMD_M_ABS.format(current), "l"]
        for point in points[1:]:
            relative = point - current
            current = point
            d.append(CMD_CONT.format(relative))
        if close:
            d.append("Z")
        return " ".join(d)

    @staticmethod
    def make_multi_line_str(lines: Sequence[tuple[Vec2, Vec2]]) -> str:
        assert len(lines) > 0
        start, end = lines[0]
        d: list[str] = [CMD_M_ABS.format(start), CMD_L_REL.format(end - start)]
        current = end
        for start, end in lines[1:]:
            d.append(CMD_M_REL.format(start - current))
            current = start
            d.append(CMD_L_REL.format(end - current))
            current = end
        return " ".join(d)

    @staticmethod
    @no_type_check
    def make_path_str(path: BkPath2d, close=False) -> str:
        d: list[str] = [CMD_M_ABS.format(path.start)]
        if len(path) == 0:
            return ""

        current = path.start
        for cmd in path.commands():
            end = cmd.end
            if cmd.type == Command.MOVE_TO:
                d.append(CMD_M_REL.format(end - current))
            elif cmd.type == Command.LINE_TO:
                d.append(CMD_L_REL.format(end - current))
            elif cmd.type == Command.CURVE3_TO:
                d.append(CMD_C3_REL.format(cmd.ctrl - current, end - current))
            elif cmd.type == Command.CURVE4_TO:
                d.append(
                    CMD_C4_REL.format(
                        cmd.ctrl1 - current, cmd.ctrl2 - current, end - current
                    )
                )
            current = end
        if close:
            d.append("Z")

        return " ".join(d)

    def configure(self, config: Configuration) -> None:
        self.lineweight_policy = config.lineweight_policy
        if config.min_lineweight:
            # config.min_lineweight in 1/300 inch!
            min_lineweight_mm = config.min_lineweight * 25.4 / 300
            self.min_lineweight = max(0.05, min_lineweight_mm)
        self.lineweight_scaling = config.lineweight_scaling

    def clear(self) -> None:
        pass

    def finalize(self) -> None:
        pass

    def enter_entity(self, entity, properties) -> None:
        pass

    def exit_entity(self, entity) -> None:
        pass


def alpha_to_opacity(alpha: str) -> float:
    # stroke-opacity: 0.0 = transparent; 1.0 = opaque
    # alpha: "00" = transparent; "ff" = opaque
    if len(alpha):
        try:
            return int(alpha, 16) / 255
        except ValueError:
            pass
    return 1.0


def map_lineweight_to_stroke_width(
    lineweight: float,
    min_stroke_width: int,
    max_stroke_width: int,
    min_lineweight=0.05,  # defined by DXF
    max_lineweight=2.11,  # defined by DXF
) -> int:
    """Map the DXF lineweight in mm to stroke-width in viewBox coordinates."""
    lineweight = max(min(lineweight, max_lineweight), min_lineweight) - min_lineweight
    factor = (max_stroke_width - min_stroke_width) / (max_lineweight - min_lineweight)
    return min_stroke_width + round(lineweight * factor)



# Copyright (c) 2020, Manfred Moitzi
# License: MIT License
import pytest
import math
from ezdxf.math import Matrix44, Vec3, Z_AXIS, arc_angle_span_deg
from ezdxf.math.transformtools import OCSTransform


def test_transform_angle_without_ocs():
    ocs = OCSTransform(Vec3(0, 0, 1), Matrix44.z_rotate(math.pi / 2))
    assert math.isclose(ocs.transform_angle(0), math.pi / 2)


def test_transform_length_without_ocs():
    ocs = OCSTransform(Z_AXIS, Matrix44.scale(2, 3, 4))
    assert math.isclose(ocs.transform_length((2, 0, 0)), 2 * 2)
    assert math.isclose(ocs.transform_length((0, 2, 0)), 2 * 3)
    assert math.isclose(ocs.transform_length((0, 0, 2)), 2 * 4)


class TestTransformThickness:
    @pytest.mark.parametrize("thickness", [-2, 0, +2])
    def test_no_transformation(self, thickness):
        ocs = OCSTransform(Z_AXIS, Matrix44())
        assert ocs.transform_thickness(thickness) == pytest.approx(thickness)

    @pytest.mark.parametrize("thickness", [-2, 0, +2])
    def test_reflection_in_x_axis(self, thickness):
        ocs = OCSTransform(Z_AXIS, Matrix44.scale(-2, 1, 1))
        assert ocs.new_ocs.uz.isclose(-Z_AXIS)  # flip extrusion vector
        assert ocs.transform_thickness(thickness) == pytest.approx(
            -thickness
        ), "thickness value should be inverted"

    @pytest.mark.parametrize("thickness", [-2, 0, +2])
    def test_reflection_in_y_axis(self, thickness):
        ocs = OCSTransform(Z_AXIS, Matrix44.scale(1, -2, 1))
        assert ocs.new_ocs.uz.isclose(-Z_AXIS)  # flip extrusion vector
        assert ocs.transform_thickness(thickness) == pytest.approx(
            -thickness
        ), "thickness value should be inverted"

    @pytest.mark.parametrize("thickness", [-2, 0, +2])
    def test_reflection_in_z_axis(self, thickness):
        ocs = OCSTransform(Z_AXIS, Matrix44.scale(1, 1, -2))
        assert ocs.new_ocs.uz.isclose(Z_AXIS)  # extrusion vector unchanged
        assert ocs.transform_thickness(thickness) == pytest.approx(
            -2 * thickness
        ), "thickness value should be -2x"

    @pytest.mark.parametrize("thickness", [-2, 0, +2])
    def test_reflection_in_x_and_y_axis(self, thickness):
        ocs = OCSTransform(Z_AXIS, Matrix44.scale(-2, -2, 1))
        assert ocs.new_ocs.uz.isclose(Z_AXIS)  # extrusion vector unchanged
        assert ocs.transform_thickness(thickness) == pytest.approx(
            thickness
        ), "thickness value should be unchanged"

    @pytest.mark.parametrize("thickness", [-2, 0, +2])
    def test_reflection_in_x_and_z_axis(self, thickness):
        ocs = OCSTransform(Z_AXIS, Matrix44.scale(-2, 1, -2))
        assert ocs.new_ocs.uz.isclose(-Z_AXIS)  # flipped extrusion vector
        assert ocs.transform_thickness(thickness) == pytest.approx(
            2 * thickness
        ), "thickness value should be 2x"

    @pytest.mark.parametrize("thickness", [-2, 0, +2])
    def test_reflection_in_y_and_z_axis(self, thickness):
        ocs = OCSTransform(Z_AXIS, Matrix44.scale(1, -2, -2))
        assert ocs.new_ocs.uz.isclose(-Z_AXIS)  # flipped extrusion vector
        assert ocs.transform_thickness(thickness) == pytest.approx(
            2 * thickness
        ), "thickness value should be 2x"

    @pytest.mark.parametrize("thickness", [-2, 0, +2])
    def test_reflection_in_x_y_and_z_axis(self, thickness):
        ocs = OCSTransform(Z_AXIS, Matrix44.scale(-2, -2, -2))
        assert ocs.new_ocs.uz.isclose(Z_AXIS)  # unchanged extrusion vector
        assert ocs.transform_thickness(thickness) == pytest.approx(
            -2 * thickness
        ), "thickness value should be -2x"


class TestTransformWidth:
    @pytest.mark.parametrize("width", [-2, 0, +2])
    def test_no_transformation(self, width):
        ocs = OCSTransform(Z_AXIS, Matrix44())
        assert ocs.transform_width(width) == pytest.approx(abs(width))

    @pytest.mark.parametrize("width", [-2, 0, +2])
    def test_uniform_scaling_for_all_axis(self, width):
        ocs = OCSTransform(Z_AXIS, Matrix44.scale(-2, -2, -2))
        assert ocs.transform_width(width) == pytest.approx(
            2 * abs(width)
        ), "width should always be >= 0"

    @pytest.mark.parametrize("width", [-2, 0, +2])
    def test_x_scaling(self, width):
        ocs = OCSTransform(Z_AXIS, Matrix44.scale(2, 1, 1))
        assert ocs.transform_width(width) == pytest.approx(
            2 * abs(width)
        ), "current implementation scales by biggest x- or y-axis factor"

    @pytest.mark.parametrize("width", [-2, 0, +2])
    def test_y_scaling(self, width):
        ocs = OCSTransform(Z_AXIS, Matrix44.scale(1, 2, 1))
        assert ocs.transform_width(width) == pytest.approx(
            2 * abs(width)
        ), "current implementation scales by biggest x- or y-axis factor"

    @pytest.mark.parametrize("width", [-2, 0, +2])
    def test_non_uniform_xy_scaling_for_x(self, width):
        ocs = OCSTransform(Z_AXIS, Matrix44.scale(-3, -2, 1))
        assert ocs.transform_width(width) == pytest.approx(
            3 * abs(width)
        ), "current implementation scales by biggest x- or y-axis factor"

    @pytest.mark.parametrize("width", [-2, 0, +2])
    def test_non_uniform_xy_scaling_for_y(self, width):
        ocs = OCSTransform(Z_AXIS, Matrix44.scale(-2, -3, 1))
        assert ocs.transform_width(width) == pytest.approx(
            3 * abs(width)
        ), "current implementation scales by biggest x- or y-axis factor"


TEST_ANGLES = [
    (0, 0),
    (0, 90),
    (0, 180),
    (0, 360),
    (0, 0),
    (0, -90),
    (0, -180),
]


def normalize_angles(s, e):
    s = s % 360.0
    e = e % 360.0
    while e < s:
        e += 360
    return s, e


class TestTransformCCWAngles:
    @pytest.mark.parametrize("s,e", TEST_ANGLES)
    def test_no_transformation(self, s, e):
        ocs = OCSTransform(Z_AXIS, Matrix44())
        assert ocs.transform_ccw_arc_angles_deg(s, e) == pytest.approx([s, e])

    @pytest.mark.parametrize("s,e", TEST_ANGLES)
    @pytest.mark.parametrize("rotation", [45, 90, 180, 270, -45, -90, -180])
    def test_rotation(self, s, e, rotation):
        ocs = OCSTransform(Z_AXIS, Matrix44.z_rotate(math.radians(rotation)))
        new_angles = normalize_angles(*ocs.transform_ccw_arc_angles_deg(s, e))
        assert new_angles == pytest.approx(
            normalize_angles(s + rotation, e + rotation)
        )

    @pytest.mark.parametrize("s,e", TEST_ANGLES)
    @pytest.mark.parametrize(
        "rotation", [45, 90, 180, 270, -45, -90, -180, -270]
    )
    @pytest.mark.parametrize(
        "sx,sy",
        [(-1, 1), (1, -1), (-1, -1)],
        ids=["x", "y", "xy"],
    )
    def test_reflections(self, s, e, rotation, sx, sy):
        m = Matrix44.chain(
            Matrix44.scale(sx, sy, 1),
            Matrix44.z_rotate(rotation),
        )
        expected_start = m.transform(Vec3.from_deg_angle(s))
        expected_end = m.transform(Vec3.from_deg_angle(e))
        expected_angle_span = arc_angle_span_deg(s, e)

        ocs = OCSTransform(Z_AXIS, m)
        new_s, new_e = ocs.transform_ccw_arc_angles_deg(s, e)
        wcs_start = ocs.new_ocs.to_wcs(Vec3.from_deg_angle(new_s))
        wcs_end = ocs.new_ocs.to_wcs(Vec3.from_deg_angle(new_e))
        assert arc_angle_span_deg(new_s, new_e) == pytest.approx(
            expected_angle_span
        )
        assert wcs_start.isclose(expected_start)
        assert wcs_end.isclose(expected_end)


if __name__ == "__main__":
    pytest.main([__file__])



#  Copyright (c) 2023, Manfred Moitzi
#  License: MIT License
import pytest

from ezdxf.fonts import fonts
from pathlib import Path


@pytest.fixture(scope="session", autouse=True)
def load_sut_font_cache():
    path = Path(__file__)
    font_folder = path.parent.parent / "fonts"
    if font_folder.exists():
        fonts.build_sut_font_manager_cache(font_folder)
    else:
        exit(666)



# Copyright (c) 2022, Manfred Moitzi
# License: MIT License

import pytest
import ezdxf


def test_set_redraw_order():
    doc = ezdxf.new()
    msp = doc.modelspace()
    p1 = msp.add_point((0, 0))
    p2 = msp.add_point((0, 0))
    p3 = msp.add_point((0, 0))
    msp.set_redraw_order([
        (p1.dxf.handle, "F"),
        (p2.dxf.handle, "E"),
        (p3.dxf.handle, "D"),
    ])

    result = list(msp.get_redraw_order())
    assert result[0] == (p1.dxf.handle, "F")
    assert result[1] == (p2.dxf.handle, "E")
    assert result[2] == (p3.dxf.handle, "D")


if __name__ == "__main__":
    pytest.main([__file__])



#  Copyright (c) 2020, Manfred Moitzi
#  License: MIT License
import os
import pytest
import ezdxf


BASEDIR = os.path.dirname(__file__)
DATADIR = "data"


@pytest.fixture(params=["small_r13.dxf", "small_r14.dxf", "no_layouts.dxf"])
def filename(request):
    filename = os.path.join(BASEDIR, DATADIR, request.param)
    if not os.path.exists(filename):
        pytest.skip("File {} not found.".format(filename))
    return filename


def test_open_R13_R14(filename, tmpdir):
    doc = ezdxf.readfile(filename)
    assert "Model" in doc.layouts, "Model space not found"
    assert "Layout1" in doc.layouts, "Paper space not found"
    assert doc.dxfversion >= "AC1015"
    msp = doc.modelspace()
    msp.add_line((0, 0), (10, 3))
    converted = str(tmpdir.join("converted_AC1015.dxf"))
    doc.saveas(converted)
    assert os.path.exists(converted)



#  Copyright (c) 2022, Manfred Moitzi
#  License: MIT License

import pytest
from ezdxf.math import intersection_line_polygon_3d, Vec3


@pytest.fixture(scope="module")
def polygon():
    return Vec3.list([(-1, -1, 1), (1, -1, 1), (1, 1, 1), (-1, 1, 1)])


def test_polygon_with_invalid_normal():
    polygon = Vec3.list([(0, 0, 0), (1, 0, 0), (2, 0, 0)])
    ip = intersection_line_polygon_3d(Vec3(0, 0, 0), Vec3(0, 0, 5), polygon)
    assert ip is None


def test_intersection_point_inside(polygon):
    ip = intersection_line_polygon_3d(Vec3(0, 0, 0), Vec3(0, 0, 5), polygon)
    assert ip.isclose((0, 0, 1))


def test_intersection_point_at_boundary_line(polygon):
    ip = intersection_line_polygon_3d(
        Vec3(1, 0, 0), Vec3(1, 0, 5), polygon, boundary=True
    )
    assert ip.isclose((1, 0, 1))


def test_ignore_intersection_point_at_boundary_line(polygon):
    ip = intersection_line_polygon_3d(
        Vec3(1, 0, 0), Vec3(1, 0, 5), polygon, boundary=False
    )
    assert ip is None


def test_ignore_coplanar_start_point_as_intersection_point(polygon):
    ip = intersection_line_polygon_3d(
        Vec3(0, 0, 1), Vec3(0, 0, 5), polygon, coplanar=False
    )
    assert ip is None


def test_ignore_coplanar_end_point_as_intersection_point(polygon):
    ip = intersection_line_polygon_3d(
        Vec3(0, 0, 0), Vec3(0, 0, 1), polygon, coplanar=False
    )
    assert ip is None


def test_intersection_point_outside():
    polygon = Vec3.list([(-1, -1, 1), (1, -1, 1), (1, 1, 1), (-1, 1, 1)])
    ip = intersection_line_polygon_3d(Vec3(2, 0, 0), Vec3(2, 0, 5), polygon)
    assert ip is None


def test_ignore_coplanar_line(polygon):
    ip = intersection_line_polygon_3d(Vec3(1, 0, 1), Vec3(2, 0, 1), polygon)
    assert ip is None


if __name__ == "__main__":
    pytest.main([__file__])



# Copyright (c) 2020-2022, Manfred Moitzi
# License: MIT License
from __future__ import annotations
from typing import Iterable, NamedTuple, BinaryIO

from .const import DXFStructureError
from ezdxf.tools.codepage import toencoding


class IndexEntry(NamedTuple):
    code: int
    value: str
    location: int
    line: int


class FileStructure:
    """DXF file structure representation stored as file locations.

    Store all DXF structure tags and some other tags as :class:`IndexEntry`
    tuples:

        - code: group code
        - value: tag value as string
        - location: file location as int
        - line: line number as int

    Indexed tags:

        - structure tags, every tag with group code 0
        - section names, (2, name) tag following a (0, SECTION) tag
        - entity handle tags with group code 5, the DIMSTYLE handle group code
          105 is also stored as group code 5

    """

    def __init__(self, filename: str):
        # stores the file system name of the DXF document.
        self.filename: str = filename
        # DXF version if header variable $ACADVER is present, default is DXFR12
        self.version: str = "AC1009"
        # Python encoding required to read the DXF document as text file.
        self.encoding: str = "cp1252"
        self.index: list[IndexEntry] = []

    def print(self) -> None:
        print(f"Filename: {self.filename}")
        print(f"DXF Version: {self.version}")
        print(f"encoding: {self.encoding}")
        for entry in self.index:
            print(f"Line: {entry.line} - ({entry.code}, {entry.value})")

    def get(self, code: int, value: str, start: int = 0) -> int:
        """Returns index of first entry matching `code` and `value`."""
        self_index = self.index
        index: int = start
        count: int = len(self_index)
        while index < count:
            entry = self_index[index]
            if entry.code == code and entry.value == value:
                return index
            index += 1
        raise ValueError(f"No entry for tag ({code}, {value}) found.")

    def fetchall(
        self, code: int, value: str, start: int = 0
    ) -> Iterable[IndexEntry]:
        """Iterate over all specified entities.

        e.g. fetchall(0, 'LINE') returns an iterator for all LINE entities.

        """
        for entry in self.index[start:]:
            if entry.code == code and entry.value == value:
                yield entry


def load(filename: str) -> FileStructure:
    """Load DXF file structure for file `filename`, the file has to be seekable.

    Args:
        filename: file system file name

    Raises:
        DXFStructureError: Invalid or incomplete DXF file.

    """
    file_structure = FileStructure(filename)
    file: BinaryIO = open(filename, mode="rb")
    line: int = 1
    eof: bool = False
    header: bool = False
    index: list[IndexEntry] = []
    prev_code: int = -1
    prev_value: bytes = b""
    structure = None  # the current structure tag: 'SECTION', 'LINE', ...

    def load_tag() -> tuple[int, bytes]:
        nonlocal line
        try:
            code = int(file.readline())
        except ValueError:
            raise DXFStructureError(f"Invalid group code in line {line}")

        if code < 0 or code > 1071:
            raise DXFStructureError(f"Invalid group code {code} in line {line}")
        value = file.readline().rstrip(b"\r\n")
        line += 2
        return code, value

    def load_header_var() -> str:
        _, value = load_tag()
        return value.decode()

    while not eof:
        location = file.tell()
        tag_line = line
        try:
            code, value = load_tag()
            if header and code == 9:
                if value == b"$ACADVER":
                    file_structure.version = load_header_var()
                elif value == b"$DWGCODEPAGE":
                    file_structure.encoding = toencoding(load_header_var())
                continue
        except IOError:
            break

        if code == 0:
            # All structure tags have group code == 0, store file location
            structure = value
            index.append(IndexEntry(0, value.decode(), location, tag_line))
            eof = value == b"EOF"

        elif code == 2 and prev_code == 0 and prev_value == b"SECTION":
            # Section name is the tag (2, name) following the (0, SECTION) tag.
            header = value == b"HEADER"
            index.append(IndexEntry(2, value.decode(), location, tag_line))

        elif code == 5 and structure != b"DIMSTYLE":
            # Entity handles have always group code 5.
            index.append(IndexEntry(5, value.decode(), location, tag_line))

        elif code == 105 and structure == b"DIMSTYLE":
            # Except the DIMSTYLE table entry has group code 105.
            index.append(IndexEntry(5, value.decode(), location, tag_line))

        prev_code = code
        prev_value = value

    file.close()
    if not eof:
        raise DXFStructureError(f"Unexpected end of file.")

    if file_structure.version >= "AC1021":  # R2007 and later
        file_structure.encoding = "utf-8"
    file_structure.index = index
    return file_structure



#  Copyright (c) 2021, Manfred Moitzi
#  License: MIT License
import pytest
from pathlib import Path
import subprocess

# Got errors at github actions for ubuntu-latest:
# Error was using (works on Windows):
#       subprocess.run("ezdxf -V")
#
# This works on Linux and Windows:
#       subprocess.run(["ezdxf", "-V"])

TEST_DATA = Path(__file__).parent / "data"
CRLF = b"\r\n"


def test_version():
    result = subprocess.run(["ezdxf", "-V"], capture_output=True)
    assert result.returncode == 0
    assert result.stdout.startswith(b"ezdxf")


def test_audit_existing_file():
    result = subprocess.run(
        ["ezdxf", "audit", str(TEST_DATA / "POLI-ALL210_12.DXF")],
        capture_output=True,
    )
    assert result.returncode == 0
    assert result.stdout.rstrip(CRLF).endswith(b"No errors found.")


def test_audit_file_not_found():
    result = subprocess.run(
        ["ezdxf", "audit", "nofile.dxf"], capture_output=True
    )
    assert result.returncode == 0
    assert result.stderr.rstrip(CRLF) == b"File(s) 'nofile.dxf' not found."


if __name__ == "__main__":
    pytest.main([__file__])



# Copyright (c) 2019-2024 Manfred Moitzi
# License: MIT License
from __future__ import annotations
from typing import TYPE_CHECKING, Optional
from typing_extensions import Self
import copy

from ezdxf.audit import Auditor, AuditError
from ezdxf.lldxf import validator
from ezdxf.math import NULLVEC, Vec3, Z_AXIS, OCS, Matrix44
from ezdxf.lldxf.attributes import (
    DXFAttr,
    DXFAttributes,
    DefSubclass,
    XType,
    RETURN_DEFAULT,
    group_code_mapping,
)
from ezdxf.lldxf import const
from ezdxf.lldxf.types import EMBEDDED_OBJ_MARKER, EMBEDDED_OBJ_STR
from ezdxf.enums import MAP_MTEXT_ALIGN_TO_FLAGS, TextHAlign, TextVAlign
from ezdxf.tools import set_flag_state
from ezdxf.tools.text import (
    load_mtext_content,
    fast_plain_mtext,
    plain_mtext,
)

from .dxfns import SubclassProcessor, DXFNamespace
from .dxfentity import base_class
from .dxfgfx import acdb_entity, elevation_to_z_axis
from .text import Text, acdb_text, acdb_text_group_codes
from .mtext import (
    acdb_mtext_group_codes,
    MText,
    export_mtext_content,
    acdb_mtext,
)
from .factory import register_entity
from .copy import default_copy

if TYPE_CHECKING:
    from ezdxf.lldxf.tagwriter import AbstractTagWriter
    from ezdxf.lldxf.tags import Tags
    from ezdxf import xref


__all__ = ["AttDef", "Attrib", "copy_attrib_as_text", "BaseAttrib"]

# Where is it valid to place an ATTRIB entity:
# - YES: attached to an INSERT entity
# - NO: stand-alone entity in model space - ignored by BricsCAD and TrueView
# - NO: stand-alone entity in paper space - ignored by BricsCAD and TrueView
# - NO: stand-alone entity in block layout - ignored by BricsCAD and TrueView
#
# The RECOVER command of BricsCAD removes the stand-alone ATTRIB entities:
# "Invalid subentity type AcDbAttribute(<handle>)"
#
# IMPORTANT: placing ATTRIB at an invalid layout does NOT create an invalid DXF file!
#
# Where is it valid to place an ATTDEF entity:
# - NO: attached to an INSERT entity
# - YES: stand-alone entity in a BLOCK layout - BricsCAD and TrueView render the
#        TAG in the block editor and does not render the ATTDEF as block content
#        for the INSERT entity.
# - YES: stand-alone entity in model space - BricsCAD and TrueView render the
#        TAG not the default text - the model space is also a block content
#        (XREF, see also INSERT entity)
# - YES: stand-alone entity in paper space - same as model space, although a
#        paper space can not be used as XREF.

# DXF Reference for ATTRIB is a total mess and incorrect, the AcDbText subclass
# for the ATTRIB entity is the same as for the TEXT entity, but the valign field
# from the 2nd AcDbText subclass of the TEXT entity is stored in the
# AcDbAttribute subclass:
attrib_fields = {
    # "version": DXFAttr(280, default=0, dxfversion=const.DXF2010),
    # The "version" tag has the same group code as the lock_position tag!!!!!
    # Version number: 0 = 2010
    # This tag is not really used (at least by BricsCAD) but there exists DXF files
    # which do use this tag: "dxftest\attrib\attrib_with_mtext_R2018.dxf"
    # ezdxf stores the last group code 280 as "lock_position" attribute and does
    # not export a version tag for any DXF version.
    # Tag string (cannot contain spaces):
    # Mandatory by AutoCAD!
    "tag": DXFAttr(
        2,
        default="",
        validator=validator.is_valid_attrib_tag,
        fixer=validator.fix_attrib_tag,
    ),
    # 1 = Attribute is invisible (does not appear)
    # 2 = This is a constant attribute
    # 4 = Verification is required on input of this attribute
    # 8 = Attribute is preset (no prompt during insertion)
    "flags": DXFAttr(70, default=0),
    # Field length (optional) (not currently used)
    "field_length": DXFAttr(73, default=0, optional=True),
    # Vertical text justification type (optional); see group code 73 in TEXT
    "valign": DXFAttr(
        74,
        default=0,
        optional=True,
        validator=validator.is_in_integer_range(0, 4),
        fixer=RETURN_DEFAULT,
    ),
    # Lock position flag. Locks the position of the attribute within the block
    # reference, example of double use of group codes in one sub class
    "lock_position": DXFAttr(
        280,
        default=0,
        dxfversion=const.DXF2007,  # tested with BricsCAD 2023/TrueView 2023
        optional=True,
        validator=validator.is_integer_bool,
        fixer=RETURN_DEFAULT,
    ),
    # Attribute type:
    # 1 = single line
    # 2 = multiline ATTRIB
    # 4 = multiline ATTDEF
    "attribute_type": DXFAttr(
        71,
        default=const.ATTRIB_TYPE_SINGLE_LINE,
        dxfversion=const.DXF2018,
        optional=True,
        validator=validator.is_one_of({1, 2, 4}),
        fixer=RETURN_DEFAULT,
    ),
}

# ATTDEF has an additional field: 'prompt'
# DXF attribute definitions are immutable, a shallow copy is sufficient:
attdef_fields = dict(attrib_fields)
attdef_fields["prompt"] = DXFAttr(
    3,
    default="",
    validator=validator.is_valid_one_line_text,
    fixer=validator.fix_one_line_text,
)

acdb_attdef = DefSubclass("AcDbAttributeDefinition", attdef_fields)
acdb_attdef_group_codes = group_code_mapping(acdb_attdef)
acdb_attrib = DefSubclass("AcDbAttribute", attrib_fields)
acdb_attrib_group_codes = group_code_mapping(acdb_attrib)

# --------------------------------------------------------------------------------------
# Does subclass AcDbXrecord really exist? Only the documentation in the DXF reference
# exists, no real world examples seen so far - it wouldn't be the first error or misleading
# information in the DXF reference.
# --------------------------------------------------------------------------------------
# For XRECORD the tag order is important and group codes appear multiple times,
# therefore this attribute definition needs a special treatment!
acdb_attdef_xrecord = DefSubclass(
    "AcDbXrecord",
    [  # type: ignore
        # Duplicate record cloning flag (determines how to merge duplicate entries):
        # 1 = Keep existing
        ("cloning", DXFAttr(280, default=1)),
        # MText flag:
        # 2 = multiline attribute
        # 4 = constant multiline attribute definition
        ("mtext_flag", DXFAttr(70, default=0)),
        # isReallyLocked flag:
        #     0 = unlocked
        #     1 = locked
        (
            "really_locked",
            DXFAttr(
                70,
                default=0,
                validator=validator.is_integer_bool,
                fixer=RETURN_DEFAULT,
            ),
        ),
        # Number of secondary attributes or attribute definitions:
        ("secondary_attribs_count", DXFAttr(70, default=0)),
        # Hard-pointer id of secondary attribute(s) or attribute definition(s):
        ("secondary_attribs_handle", DXFAttr(340, default="0")),
        # Alignment point of attribute or attribute definition:
        ("align_point", DXFAttr(10, xtype=XType.point3d, default=NULLVEC)),
        ("current_annotation_scale", DXFAttr(40, default=0)),
        # attribute or attribute definition tag string
        (
            "tag",
            DXFAttr(
                2,
                default="",
                validator=validator.is_valid_attrib_tag,
                fixer=validator.fix_attrib_tag,
            ),
        ),
    ],
)


# Just for documentation:
# The "attached" MTEXT feature most likely does not exist!
#
#   A special MTEXT entity can follow the ATTDEF and ATTRIB entity, which starts
#   as a usual DXF entity with (0, 'MTEXT'), so processing can't be done here,
#   because for ezdxf is this a separated Entity.
#
#   The attached MTEXT entity: owner is None and handle is None
#   Linked as attribute `attached_mtext`.
#   I don't have seen this combination of entities in real world examples and is
#   ignored by ezdxf for now.
#
# No DXF files available which uses this feature - misleading DXF Reference!?

# Attrib and Attdef can have embedded MTEXT entities located in the
# <Embedded Object> subclass, see issue #258


class BaseAttrib(Text):
    XRECORD_DEF = acdb_attdef_xrecord

    def __init__(self) -> None:
        super().__init__()
        # Does subclass AcDbXrecord really exist?
        self._xrecord: Optional[Tags] = None
        self._embedded_mtext: Optional[EmbeddedMText] = None

    def copy_data(self, entity: Self, copy_strategy=default_copy) -> None:
        """Copy entity data, xrecord data and embedded MTEXT are not stored
        in the entity database.
        """
        assert isinstance(entity, BaseAttrib)
        entity._xrecord = copy.deepcopy(self._xrecord)
        entity._embedded_mtext = copy.deepcopy(self._embedded_mtext)

    def load_embedded_mtext(self, processor: SubclassProcessor) -> None:
        if not processor.embedded_objects:
            return
        embedded_object = processor.embedded_objects[0]
        if embedded_object:
            mtext = EmbeddedMText()
            mtext.load_dxf_tags(processor)
            self._embedded_mtext = mtext

    def export_dxf_r2018_features(self, tagwriter: AbstractTagWriter) -> None:
        tagwriter.write_tag2(71, self.dxf.attribute_type)
        tagwriter.write_tag2(72, 0)  # unknown tag
        if self.dxf.hasattr("align_point"):
            # duplicate align point - why?
            tagwriter.write_vertex(11, self.dxf.align_point)

        if self._xrecord:
            tagwriter.write_tags(self._xrecord)
        if self._embedded_mtext:
            self._embedded_mtext.export_dxf_tags(tagwriter)

    @property
    def is_const(self) -> bool:
        """This is a constant attribute if ``True``."""
        return bool(self.dxf.flags & const.ATTRIB_CONST)

    @is_const.setter
    def is_const(self, state: bool) -> None:
        self.dxf.flags = set_flag_state(self.dxf.flags, const.ATTRIB_CONST, state)

    @property
    def is_invisible(self) -> bool:
        """Attribute is invisible if ``True``."""
        return bool(self.dxf.flags & const.ATTRIB_INVISIBLE)

    @is_invisible.setter
    def is_invisible(self, state: bool) -> None:
        self.dxf.flags = set_flag_state(self.dxf.flags, const.ATTRIB_INVISIBLE, state)

    @property
    def is_verify(self) -> bool:
        """Verification is required on input of this attribute. (interactive CAD
        application feature)
        """
        return bool(self.dxf.flags & const.ATTRIB_VERIFY)

    @is_verify.setter
    def is_verify(self, state: bool) -> None:
        self.dxf.flags = set_flag_state(self.dxf.flags, const.ATTRIB_VERIFY, state)

    @property
    def is_preset(self) -> bool:
        """No prompt during insertion. (interactive CAD application feature)"""
        return bool(self.dxf.flags & const.ATTRIB_IS_PRESET)

    @is_preset.setter
    def is_preset(self, state: bool) -> None:
        self.dxf.flags = set_flag_state(self.dxf.flags, const.ATTRIB_IS_PRESET, state)

    @property
    def has_embedded_mtext_entity(self) -> bool:
        """Returns ``True`` if the entity has an embedded MTEXT entity for multi-line
        support.
        """
        return bool(self._embedded_mtext)

    def virtual_mtext_entity(self) -> MText:
        """Returns the embedded MTEXT entity as a regular but virtual
        :class:`MText` entity with the same graphical properties as the
        host entity.
        """
        if not self._embedded_mtext:
            raise TypeError("no embedded MTEXT entity exist")
        mtext = self._embedded_mtext.virtual_mtext_entity()
        mtext.update_dxf_attribs(self.graphic_properties())
        return mtext

    def plain_mtext(self, fast=True) -> str:
        """Returns the embedded MTEXT content without formatting codes.
        Returns an empty string if no embedded MTEXT entity exist.

        The `fast` mode is accurate if the DXF content was created by
        reliable (and newer) CAD applications like AutoCAD or BricsCAD.
        The `accurate` mode is for some rare cases where the content was
        created by older CAD applications or unreliable DXF libraries and CAD
        applications.

        The `accurate` mode is **much** slower than the `fast` mode.

        Args:
            fast: uses the `fast` mode to extract the plain MTEXT content if
                ``True`` or the `accurate` mode if set to ``False``

        """
        if self._embedded_mtext:
            text = self._embedded_mtext.text
            if fast:
                return fast_plain_mtext(text, split=False)  # type: ignore
            else:
                return plain_mtext(text, split=False)  # type: ignore
        return ""

    def set_mtext(self, mtext: MText, graphic_properties=True) -> None:
        """Set multi-line properties from a :class:`MText` entity.

        The multi-line ATTRIB/ATTDEF entity requires DXF R2018, otherwise an
        ordinary single line ATTRIB/ATTDEF entity will be exported.

        Args:
            mtext: source :class:`MText` entity
            graphic_properties: copy graphic properties (color, layer, ...) from
                source MTEXT if ``True``

        """
        if self._embedded_mtext is None:
            self._embedded_mtext = EmbeddedMText()
        self._embedded_mtext.set_mtext(mtext)
        _update_content_from_mtext(self, mtext)
        _update_location_from_mtext(self, mtext)

        # set attribute type:
        if isinstance(self, Attrib):
            attribute_type = const.ATTRIB_TYPE_MULTI_LINE
        else:
            attribute_type = const.ATTDEF_TYPE_MULTI_LINE
        self.dxf.attribute_type = attribute_type

        # misc properties
        self.dxf.style = mtext.dxf.style
        self.dxf.height = mtext.dxf.char_height
        self.dxf.discard("width")  # controlled in MTEXT by inline codes!
        self.dxf.discard("oblique")  # controlled in MTEXT by inline codes!
        self.dxf.discard("text_generation_flag")
        if graphic_properties:
            self.update_dxf_attribs(mtext.graphic_properties())

    def embed_mtext(self, mtext: MText, graphic_properties=True) -> None:
        """Set multi-line properties from a :class:`MText` entity and destroy the
        source entity afterward.

        The multi-line ATTRIB/ATTDEF entity requires DXF R2018, otherwise an
        ordinary single line ATTRIB/ATTDEF entity will be exported.

        Args:
            mtext: source :class:`MText` entity
            graphic_properties: copy graphic properties (color, layer, ...) from
                source MTEXT if ``True``

        """
        self.set_mtext(mtext, graphic_properties)
        mtext.destroy()

    def discard_mtext(self) -> None:
        """Discard multi-line feature.

        The embedded MTEXT will be removed and the ATTRIB/ATTDEF will be converted to a
        single-line attribute.
        """
        self._embedded_mtext = None
        self.dxf.attribute_type = const.ATTRIB_TYPE_SINGLE_LINE

    def register_resources(self, registry: xref.Registry) -> None:
        """Register required resources to the resource registry."""
        super().register_resources(registry)
        if self._embedded_mtext:
            self._embedded_mtext.register_resources(registry)

    def map_resources(self, clone: Self, mapping: xref.ResourceMapper) -> None:
        """Translate resources from self to the copied entity."""
        assert isinstance(clone, BaseAttrib)
        super().map_resources(clone, mapping)
        if self._embedded_mtext and clone._embedded_mtext:
            self._embedded_mtext.map_resources(clone._embedded_mtext, mapping)
        # todo: map handles in embedded XRECORD if a real world example shows up

    def transform(self, m: Matrix44) -> Self:
        if self._embedded_mtext is None:
            super().transform(m)
        else:
            mtext = self._embedded_mtext.virtual_mtext_entity()
            mtext.transform(m)
            self.set_mtext(mtext, graphic_properties=False)
            self.post_transform(m)
        return self

    def audit(self, auditor: Auditor) -> None:
        """Validity check."""
        super().audit(auditor)
        if not self.dxf.hasattr("tag"):
            auditor.fixed_error(
                code=AuditError.TAG_ATTRIBUTE_MISSING,
                message=f'Missing mandatory "tag" attribute, entity {str(self)} deleted.',
            )
            auditor.trash(self)


def _update_content_from_mtext(text: Text, mtext: MText) -> None:
    content = mtext.plain_text(split=True, fast=True)
    if content:
        # In contrast to AutoCAD, just set the first line as single line
        # ATTRIB content. AutoCAD concatenates all lines into a single
        # "Line1\PLine2\P...", which (imho) is not very useful.
        text.dxf.text = content[0]


def _update_location_from_mtext(text: Text, mtext: MText) -> None:
    # TEXT is an OCS entity, MTEXT is a WCS entity
    dxf = text.dxf
    insert = Vec3(mtext.dxf.insert)
    extrusion = Vec3(mtext.dxf.extrusion)
    text_direction = mtext.get_text_direction()
    if extrusion.isclose(Z_AXIS):  # most common case
        dxf.rotation = text_direction.angle_deg
    else:
        ocs = OCS(extrusion)
        insert = ocs.from_wcs(insert)
        dxf.extrusion = extrusion.normalize()
        dxf.rotation = ocs.from_wcs(text_direction).angle_deg

    dxf.insert = insert
    dxf.align_point = insert  # the same point for all MTEXT alignments!
    dxf.halign, dxf.valign = MAP_MTEXT_ALIGN_TO_FLAGS.get(
        mtext.dxf.attachment_point, (TextHAlign.LEFT, TextVAlign.TOP)
    )


@register_entity
class AttDef(BaseAttrib):
    """DXF ATTDEF entity"""

    DXFTYPE = "ATTDEF"
    # Don't add acdb_attdef_xrecord here:
    DXFATTRIBS = DXFAttributes(base_class, acdb_entity, acdb_text, acdb_attdef)

    def load_dxf_attribs(
        self, processor: Optional[SubclassProcessor] = None
    ) -> DXFNamespace:
        dxf = super(Text, self).load_dxf_attribs(processor)
        # Do not call Text loader.
        if processor:
            processor.fast_load_dxfattribs(dxf, acdb_text_group_codes, 2, recover=True)
            processor.fast_load_dxfattribs(
                dxf, acdb_attdef_group_codes, 3, recover=True
            )
            self._xrecord = processor.find_subclass(self.XRECORD_DEF.name)  # type: ignore
            self.load_embedded_mtext(processor)
            if processor.r12:
                # Transform elevation attribute from R11 to z-axis values:
                elevation_to_z_axis(dxf, ("insert", "align_point"))

        return dxf

    def export_entity(self, tagwriter: AbstractTagWriter) -> None:
        # Text() writes 2x AcDbText which is not suitable for AttDef()
        self.export_acdb_entity(tagwriter)
        self.export_acdb_text(tagwriter)
        self.export_acdb_attdef(tagwriter)
        if tagwriter.dxfversion >= const.DXF2018:
            self.export_dxf_r2018_features(tagwriter)

    def export_acdb_attdef(self, tagwriter: AbstractTagWriter) -> None:
        if tagwriter.dxfversion > const.DXF12:
            tagwriter.write_tag2(const.SUBCLASS_MARKER, acdb_attdef.name)
        self.dxf.export_dxf_attribs(
            tagwriter,
            [
                # write version tag (280, 0) here, if required in the future
                "prompt",
                "tag",
                "flags",
                "field_length",
                "valign",
                "lock_position",
            ],
        )


@register_entity
class Attrib(BaseAttrib):
    """DXF ATTRIB entity"""

    DXFTYPE = "ATTRIB"
    # Don't add acdb_attdef_xrecord here:
    DXFATTRIBS = DXFAttributes(base_class, acdb_entity, acdb_text, acdb_attrib)

    def load_dxf_attribs(
        self, processor: Optional[SubclassProcessor] = None
    ) -> DXFNamespace:
        dxf = super(Text, self).load_dxf_attribs(processor)
        # Do not call Text loader.
        if processor:
            processor.fast_load_dxfattribs(dxf, acdb_text_group_codes, 2, recover=True)
            processor.fast_load_dxfattribs(
                dxf, acdb_attrib_group_codes, 3, recover=True
            )
            self._xrecord = processor.find_subclass(self.XRECORD_DEF.name)  # type: ignore
            self.load_embedded_mtext(processor)
            if processor.r12:
                # Transform elevation attribute from R11 to z-axis values:
                elevation_to_z_axis(dxf, ("insert", "align_point"))
        return dxf

    def export_entity(self, tagwriter: AbstractTagWriter) -> None:
        # Text() writes 2x AcDbText which is not suitable for AttDef()
        self.export_acdb_entity(tagwriter)
        self.export_acdb_attrib_text(tagwriter)
        self.export_acdb_attrib(tagwriter)
        if tagwriter.dxfversion >= const.DXF2018:
            self.export_dxf_r2018_features(tagwriter)

    def export_acdb_attrib_text(self, tagwriter: AbstractTagWriter) -> None:
        # Despite the similarities to TEXT, it is different to
        # Text.export_acdb_text():
        if tagwriter.dxfversion > const.DXF12:
            tagwriter.write_tag2(const.SUBCLASS_MARKER, acdb_text.name)
        self.dxf.export_dxf_attribs(
            tagwriter,
            [
                "insert",
                "height",
                "text",
                "thickness",
                "rotation",
                "oblique",
                "style",
                "width",
                "halign",
                "align_point",
                "text_generation_flag",
                "extrusion",
            ],
        )

    def export_acdb_attrib(self, tagwriter: AbstractTagWriter) -> None:
        if tagwriter.dxfversion > const.DXF12:
            tagwriter.write_tag2(const.SUBCLASS_MARKER, acdb_attrib.name)
        self.dxf.export_dxf_attribs(
            tagwriter,
            [
                # write version tag (280, 0) here, if required in the future
                "tag",
                "flags",
                "field_length",
                "valign",
                "lock_position",
            ],
        )


IGNORE_FROM_ATTRIB = {
    "handle",
    "owner",
    "version",
    "prompt",
    "tag",
    "flags",
    "field_length",
    "lock_position",
}


def copy_attrib_as_text(attrib: BaseAttrib):
    """Returns the content of the ATTRIB/ATTDEF entity as a new virtual TEXT or
    MTEXT entity.

    """
    if attrib.has_embedded_mtext_entity:
        return attrib.virtual_mtext_entity()
    dxfattribs = attrib.dxfattribs(drop=IGNORE_FROM_ATTRIB)
    return Text.new(dxfattribs=dxfattribs, doc=attrib.doc)


class EmbeddedMTextNS(DXFNamespace):
    _DXFATTRIBS = DXFAttributes(acdb_mtext)

    @property
    def dxfattribs(self) -> DXFAttributes:
        return self._DXFATTRIBS

    @property
    def dxftype(self) -> str:
        return "Embedded MText"


class EmbeddedMText:
    """Representation of the embedded MTEXT object in ATTRIB and ATTDEF.

    Introduced in DXF R2018? The DXF reference of the `MTEXT`_ entity
    documents only the attached MTEXT entity. The ODA DWG specs includes all
    MTEXT attributes of MTEXT starting at group code 10

    Stores the required parameters to be shown as as MTEXT.
    The AcDbText subclass contains  the first line of the embedded MTEXT as
    plain text content as group code 1, but this tag seems not to be maintained
    if the ATTRIB entity is copied.

    Some DXF attributes are duplicated and maintained by the CAD application:

        - textstyle: same group code 7 (AcDbText, EmbeddedObject)
        - text (char) height: same group code 40 (AcDbText, EmbeddedObject)

    .. _MTEXT: https://help.autodesk.com/view/OARX/2018/ENU/?guid=GUID-7DD8B495-C3F8-48CD-A766-14F9D7D0DD9B

    """

    def __init__(self) -> None:
        # Attribute "dxf" contains the DXF attributes defined in subclass
        # "AcDbMText"
        self.dxf = EmbeddedMTextNS()
        self.text: str = ""

    def copy(self) -> EmbeddedMText:
        copy_ = EmbeddedMText()
        copy_.dxf = copy.deepcopy(self.dxf)
        return copy_

    __copy__ = copy

    def load_dxf_tags(self, processor: SubclassProcessor) -> None:
        tags = processor.fast_load_dxfattribs(
            self.dxf,
            group_code_mapping=acdb_mtext_group_codes,
            subclass=processor.embedded_objects[0],
            recover=False,
        )
        self.text = load_mtext_content(tags)

    def virtual_mtext_entity(self) -> MText:
        """Returns the embedded MTEXT entity as regular but virtual MTEXT
        entity. This entity does not have the graphical attributes of the host
        entity (ATTRIB/ATTDEF).

        """
        mtext = MText.new(dxfattribs=self.dxf.all_existing_dxf_attribs())
        mtext.text = self.text
        return mtext

    def set_mtext(self, mtext: MText) -> None:
        """Set embedded MTEXT attributes from given `mtext` entity."""
        self.text = mtext.text
        dxf = self.dxf
        for k, v in mtext.dxf.all_existing_dxf_attribs().items():
            if dxf.is_supported(k):
                dxf.set(k, v)

    def set_required_dxf_attributes(self):
        # These attributes are always present in DXF files created by Autocad:
        dxf = self.dxf
        for key, default in (
            ("insert", NULLVEC),
            ("char_height", 2.5),
            ("width", 0.0),
            ("defined_height", 0.0),
            ("attachment_point", 1),
            ("flow_direction", 5),
            ("style", "Standard"),
            ("line_spacing_style", 1),
            ("line_spacing_factor", 1.0),
        ):
            if not dxf.hasattr(key):
                dxf.set(key, default)

    def export_dxf_tags(self, tagwriter: AbstractTagWriter) -> None:
        """Export embedded MTEXT as "Embedded Object"."""
        tagwriter.write_tag2(EMBEDDED_OBJ_MARKER, EMBEDDED_OBJ_STR)
        self.set_required_dxf_attributes()
        self.dxf.export_dxf_attribs(
            tagwriter,
            [
                "insert",
                "char_height",
                "width",
                "defined_height",
                "attachment_point",
                "flow_direction",
            ],
        )
        export_mtext_content(self.text, tagwriter)
        self.dxf.export_dxf_attribs(
            tagwriter,
            [
                "style",
                "extrusion",
                "text_direction",
                "rect_width",
                "rect_height",
                "rotation",
                "line_spacing_style",
                "line_spacing_factor",
                "box_fill_scale",
                "bg_fill",
                "bg_fill_color",
                "bg_fill_true_color",
                "bg_fill_color_name",
                "bg_fill_transparency",
            ],
        )

    def register_resources(self, registry: xref.Registry) -> None:
        """Register required resources to the resource registry."""
        if self.dxf.hasattr("style"):
            registry.add_text_style(self.dxf.style)

    def map_resources(self, clone: EmbeddedMText, mapping: xref.ResourceMapper) -> None:
        """Translate resources from self to the copied entity."""
        if clone.dxf.hasattr("style"):
            clone.dxf.style = mapping.get_text_style(clone.dxf.style)



# Copyright (c) 2019-2021 Manfred Moitzi
# License: MIT License
import pytest

from ezdxf.entities.attrib import AttDef
from ezdxf.lldxf import const
from ezdxf.lldxf.tagwriter import TagCollector, basic_tags_from_text


TEST_CLASS = AttDef
TEST_TYPE = "ATTDEF"

ENTITY_R12 = """0
ATTDEF
5
0
8
0
10
0.0
20
0.0
30
0.0
40
1.0
1
DEFAULTTEXT
50
0.0
51
0.0
7
STANDARD
41
1.0
71
0
72
0
11
0.0
21
0.0
31
0.0
3
PROMPTTEXT
2
TAG
70
0
74
0
"""

ENTITY_R2000 = """0
ATTDEF
5
0
330
0
100
AcDbEntity
8
0
100
AcDbText
10
0.0
20
0.0
30
0.0
40
1.0
1
DEFAULTTEXT
50
0.0
51
0.0
7
STANDARD
41
1.0
71
0
72
0
11
0.0
21
0.0
31
0.0
100
AcDbAttributeDefinition
3
PROMPTTEXT
2
TAG
70
0
73
0
74
0
"""


@pytest.fixture(params=[ENTITY_R12, ENTITY_R2000])
def entity(request):
    return TEST_CLASS.from_text(request.param)


def test_registered():
    from ezdxf.entities.factory import ENTITY_CLASSES

    assert TEST_TYPE in ENTITY_CLASSES


def test_default_init():
    entity = TEST_CLASS()
    assert entity.dxftype() == TEST_TYPE


def test_default_new():
    entity = TEST_CLASS.new(
        handle="ABBA",
        owner="0",
        dxfattribs={
            "color": "7",
            "insert": (1, 2, 3),
        },
    )
    assert entity.dxf.layer == "0"
    assert entity.dxf.color == 7
    assert entity.dxf.linetype == "BYLAYER"
    assert entity.dxf.insert == (1, 2, 3)
    assert entity.dxf.insert.x == 1, "is not Vec3 compatible"
    assert entity.dxf.insert.y == 2, "is not Vec3 compatible"
    assert entity.dxf.insert.z == 3, "is not Vec3 compatible"
    # can set DXF R2007 value
    entity.dxf.shadow_mode = 1
    assert entity.dxf.shadow_mode == 1
    assert entity.dxf.extrusion == (0.0, 0.0, 1.0)
    assert entity.dxf.hasattr("extrusion") is False, "just the default value"


def test_load_from_text(entity):
    assert entity.dxf.layer == "0"
    assert entity.dxf.color == 256, "default color is 256 (by layer)"
    assert entity.dxf.insert == (0, 0, 0)


@pytest.mark.parametrize(
    "txt,ver", [(ENTITY_R2000, const.DXF2000), (ENTITY_R12, const.DXF12)]
)
def test_write_dxf(txt, ver):
    expected = basic_tags_from_text(txt)
    attdef = TEST_CLASS.from_text(txt)
    collector = TagCollector(dxfversion=ver, optional=True)
    attdef.export_dxf(collector)
    assert collector.tags == expected

    collector2 = TagCollector(dxfversion=ver, optional=False)
    attdef.export_dxf(collector2)
    assert collector.has_all_tags(collector2)


class TestEmbeddedMTextSupport:
    @pytest.fixture
    def attdef(self) -> AttDef:
        return AttDef.from_text(EMBEDDED_MTEXT)

    def test_has_embedded_mtext(self, attdef):
        assert attdef.has_embedded_mtext_entity is True

    def test_get_plain_mtext(self, attdef):
        assert attdef.plain_mtext() == "TEST VENUE\nTEST FLOOR PLAN"

    def test_get_virtual_mtext_entity(self, attdef):
        mtext = attdef.virtual_mtext_entity()
        assert mtext.plain_text() == "TEST VENUE\nTEST FLOOR PLAN"

    def test_attdef_graphic_attributes(self, attdef):
        assert attdef.dxf.color == 7
        assert attdef.dxf.layer == "AttribLayer"

    def test_mtext_graphic_attdefutes_inherited_from_host(self, attdef):
        mtext = attdef.virtual_mtext_entity()
        assert mtext.dxf.color == 7
        assert mtext.dxf.layer == "AttribLayer"

    def test_mtext_entity_attributes(self, attdef):
        mtext = attdef.virtual_mtext_entity()
        # These seems to be the required DXF tag for the embedded MTEXT entity:
        assert mtext.dxf.insert.isclose((45.3, 45.0, 0))
        assert mtext.dxf.char_height == 3.0
        assert mtext.dxf.width == 0
        assert mtext.dxf.defined_height == 0
        assert mtext.dxf.attachment_point == 5
        assert mtext.dxf.flow_direction == 5
        assert mtext.dxf.style == "Arial_3 NARROW"
        assert mtext.dxf.line_spacing_style == 1
        assert mtext.dxf.line_spacing_factor == 1.0

    def test_dxf_export_matches_test_data(self, attdef):
        result = TagCollector.dxftags(attdef, dxfversion=const.DXF2018)
        expected = basic_tags_from_text(EMBEDDED_MTEXT)
        assert result == expected


def test_lock_position_and_ignore_version_tag():
    """Version tag and lock_position tags are present."""
    attdef = AttDef.from_text(LOCK_POSITION_AND_VERSION)
    assert attdef.dxf.lock_position == 2


def test_lock_position_without_version_tag():
    """Just the lock_position tag is present."""
    attdef = AttDef.from_text(LOCK_POSITION_WITHOUT_VERSION)
    assert attdef.dxf.lock_position == 2


def test_version_without_lock_position():
    """Only the version tag is present and the lock_position tag is missing.

    The version tag is always 0 and just for testing purposes set to 7:
    """
    attdef = AttDef.from_text(VERSION_WITHOUT_LOCK_POSITION)
    assert attdef.dxf.lock_position == 7


LOCK_POSITION_AND_VERSION = """  0
ATTDEF
5
0
330
0
100
AcDbEntity
8
AttribLayer
100
AcDbText
10
1
20
2
30
0
40
3.0
1
TEST VENUE
7
Arial_3 NARROW
100
AcDbAttributeDefinition
280
0
3
TITLE-OF-DRAWING
2
DRAWING-NAME
70
0
74
2
280
2
"""

LOCK_POSITION_WITHOUT_VERSION = """  0
ATTDEF
5
0
330
0
100
AcDbEntity
8
AttribLayer
100
AcDbText
10
1
20
2
30
0
40
3.0
1
TEST VENUE
7
Arial_3 NARROW
100
AcDbAttributeDefinition
3
TITLE-OF-DRAWING
2
DRAWING-NAME
70
0
74
2
280
2
"""

VERSION_WITHOUT_LOCK_POSITION = """  0
ATTDEF
5
0
330
0
100
AcDbEntity
8
AttribLayer
100
AcDbText
10
1
20
2
30
0
40
3.0
1
TEST VENUE
7
Arial_3 NARROW
100
AcDbAttributeDefinition
280
7
3
TITLE-OF-DRAWING
2
DRAWING-NAME
70
0
74
2
"""

EMBEDDED_MTEXT = r"""  0
ATTDEF
5
28A
330
285
100
AcDbEntity
8
AttribLayer
62
7
100
AcDbText
10
45.3
20
43.5
30
0.0
40
3.0
1
TEST VENUE
7
Arial_3 NARROW
72
1
11
45.3
21
45.0
31
0.0
100
AcDbAttributeDefinition
3
TITLE-OF-DRAWING
2
DRAWING-NAME
70
0
74
2
280
0
71
4
72
0
11
45.3
21
45.0
31
0.0
101
Embedded Object
10
45.3
20
45.0
30
0.0
40
3.0
41
0.0
46
0.0
71
5
72
5
1
TEST VENUE\PTEST FLOOR PLAN
7
Arial_3 NARROW
73
1
44
1.0
1001
AcadAnnotative
1000
AnnotativeData
1002
{
1070
1
1070
0
1002
}
"""



# Copyright (c) 2019-2022, Manfred Moitzi
# License: MIT License
from __future__ import annotations
from typing import TYPE_CHECKING, Optional
import logging
from ezdxf.lldxf import validator, const
from ezdxf.lldxf.attributes import (
    DXFAttr,
    DXFAttributes,
    DefSubclass,
    RETURN_DEFAULT,
    group_code_mapping,
)
from ezdxf.lldxf.const import DXF12, SUBCLASS_MARKER
from ezdxf.entities.dxfentity import base_class, SubclassProcessor, DXFEntity
from ezdxf.entities.layer import acdb_symbol_table_record
from .factory import register_entity

if TYPE_CHECKING:
    from ezdxf.entities import DXFNamespace
    from ezdxf.lldxf.tagwriter import AbstractTagWriter
    from ezdxf.fonts import fonts

__all__ = ["Textstyle", "get_textstyle"]
logger = logging.getLogger("ezdxf")

acdb_style = DefSubclass(
    "AcDbTextStyleTableRecord",
    {
        "name": DXFAttr(
            2,
            default="Standard",
            validator=validator.is_valid_table_name,
        ),
        # Flags: Standard flag values (bit-coded values):
        # 1 = If set, this entry describes a shape
        # 4 = Vertical text
        # 16 = If set, table entry is externally dependent on a xref
        # 32 = If both this bit and bit 16 are set, the externally dependent xref ...
        # 64 = If set, the table entry was referenced by at least one entity in ...
        # Vertical text works only for SHX fonts in AutoCAD and BricsCAD
        "flags": DXFAttr(70, default=0),
        # Fixed height, 0 if not fixed
        "height": DXFAttr(
            40,
            default=0,
            validator=validator.is_greater_or_equal_zero,
            fixer=RETURN_DEFAULT,
        ),
        # Width factor:  a.k.a. "Stretch"
        "width": DXFAttr(
            41,
            default=1,
            validator=validator.is_greater_zero,
            fixer=RETURN_DEFAULT,
        ),
        # Oblique angle in degree, 0 = vertical
        "oblique": DXFAttr(50, default=0),
        # Generation flags:
        # 2 = backward
        # 4 = mirrored in Y
        "generation_flags": DXFAttr(71, default=0),
        # Last height used:
        "last_height": DXFAttr(42, default=2.5),
        # Primary font file name:
        # ATTENTION: The font file name can be an empty string and the font family
        # may be stored in XDATA! See also posts at the (unrelated) issue #380.
        "font": DXFAttr(3, default=const.DEFAULT_TEXT_FONT),
        # Big font name, blank if none
        "bigfont": DXFAttr(4, default=""),
    },
)
acdb_style_group_codes = group_code_mapping(acdb_style)


# XDATA: This is not a reliable source for font data!
# 1001 <ctrl> ACAD
# 1000 <str> Arial  ; font-family sometimes an empty string!
# 1071 <int> 34  ; flags
# ----
# "Arial" "normal" flags = 34               = 0b00:00000000:00000000:00100010
# "Arial" "italic" flags = 16777250         = 0b01:00000000:00000000:00100010
# "Arial" "bold" flags = 33554466           = 0b10:00000000:00000000:00100010
# "Arial" "bold+italic" flags = 50331682    = 0b11:00000000:00000000:00100010


@register_entity
class Textstyle(DXFEntity):
    """DXF STYLE entity"""

    DXFTYPE = "STYLE"
    DXFATTRIBS = DXFAttributes(base_class, acdb_symbol_table_record, acdb_style)
    ITALIC = 0b01000000000000000000000000
    BOLD = 0b10000000000000000000000000

    def load_dxf_attribs(
        self, processor: Optional[SubclassProcessor] = None
    ) -> DXFNamespace:
        dxf = super().load_dxf_attribs(processor)
        if processor:
            processor.simple_dxfattribs_loader(dxf, acdb_style_group_codes)  # type: ignore
        return dxf

    def export_entity(self, tagwriter: AbstractTagWriter) -> None:
        super().export_entity(tagwriter)
        if tagwriter.dxfversion > DXF12:
            tagwriter.write_tag2(SUBCLASS_MARKER, acdb_symbol_table_record.name)
            tagwriter.write_tag2(SUBCLASS_MARKER, acdb_style.name)

        self.dxf.export_dxf_attribs(
            tagwriter,
            [
                "name",
                "flags",
                "height",
                "width",
                "oblique",
                "generation_flags",
                "last_height",
                "font",
                "bigfont",
            ],
        )

    @property
    def has_extended_font_data(self) -> bool:
        """Returns ``True`` if extended font data is present."""
        return self.has_xdata("ACAD")

    def get_extended_font_data(self) -> tuple[str, bool, bool]:
        """Returns extended font data as tuple (font-family, italic-flag,
        bold-flag).

        The extended font data is optional and not reliable! Returns
        ("", ``False``, ``False``) if extended font data is not present.

        """
        family = ""
        italic = False
        bold = False
        try:
            xdata = self.get_xdata("ACAD")
        except const.DXFValueError:
            pass
        else:
            if len(xdata) > 1:
                group_code, value = xdata[0]
                if group_code == 1000:
                    family = value
                group_code, value = xdata[1]
                if group_code == 1071:
                    italic = bool(self.ITALIC & value)
                    bold = bool(self.BOLD & value)
        return family, italic, bold

    def set_extended_font_data(
        self, family: str = "", *, italic=False, bold=False
    ) -> None:
        """Set extended font data, the font-family name `family` is not
        validated by `ezdxf`. Overwrites existing data.
        """
        if self.has_xdata("ACAD"):
            self.discard_xdata("ACAD")

        flags = 34  # unknown default flags
        if italic:
            flags += self.ITALIC
        if bold:
            flags += self.BOLD
        self.set_xdata("ACAD", [(1000, family), (1071, flags)])

    def discard_extended_font_data(self):
        """Discard extended font data."""
        self.discard_xdata("ACAD")

    @property
    def is_backward(self) -> bool:
        """Get/set text generation flag BACKWARDS, for mirrored text along the
        x-axis.
        """
        return self.get_flag_state(const.BACKWARD, "generation_flags")

    @is_backward.setter
    def is_backward(self, state) -> None:
        self.set_flag_state(const.BACKWARD, state, "generation_flags")

    @property
    def is_upside_down(self) -> bool:
        """Get/set text generation flag UPSIDE_DOWN, for mirrored text along
        the y-axis.

        """
        return self.get_flag_state(const.UPSIDE_DOWN, "generation_flags")

    @is_upside_down.setter
    def is_upside_down(self, state) -> None:
        self.set_flag_state(const.UPSIDE_DOWN, state, "generation_flags")

    @property
    def is_vertical_stacked(self) -> bool:
        """Get/set style flag VERTICAL_STACKED, for vertical stacked text."""
        return self.get_flag_state(const.VERTICAL_STACKED, "flags")

    @is_vertical_stacked.setter
    def is_vertical_stacked(self, state) -> None:
        self.set_flag_state(const.VERTICAL_STACKED, state, "flags")

    @property
    def is_shape_file(self) -> bool:
        """``True`` if entry describes a shape."""
        return self.dxf.name == "" and bool(self.dxf.flags & 1)

    def make_font(
        self,
        cap_height: Optional[float] = None,
        width_factor: Optional[float] = None,
    ) -> fonts.AbstractFont:
        """Returns a font abstraction :class:`~ezdxf.tools.fonts.AbstractFont`
        for this text style. Returns a font for a cap height of 1, if the
        text style has auto height (:attr:`Textstyle.dxf.height` is 0) and
        the given `cap_height` is ``None`` or 0.
        Uses the :attr:`Textstyle.dxf.width` attribute if the given `width_factor`
        is ``None`` or 0, the default value is 1.
        The attribute :attr:`Textstyle.dxf.big_font` is ignored.
        """
        from ezdxf.fonts import fonts

        ttf = ""
        if self.has_extended_font_data:
            family, italic, bold = self.get_extended_font_data()
            if family:
                text_style = "Italic" if italic else "Regular"
                text_weight = 700 if bold else 400
                font_face = fonts.FontFace(
                    family=family, style=text_style, weight=text_weight
                )
                ttf = fonts.find_font_file_name(font_face)
        else:
            ttf = self.dxf.get("font", const.DEFAULT_TTF)
        if ttf == "":
            ttf = const.DEFAULT_TTF
        if cap_height is None or cap_height == 0.0:
            cap_height = self.dxf.height
        if cap_height == 0.0:
            cap_height = 1.0
        if width_factor is None or width_factor == 0.0:
            width_factor = self.dxf.width
        return fonts.make_font(ttf, cap_height, width_factor)  # type: ignore


def get_textstyle(entity: DXFEntity) -> Textstyle:
    """Returns the :class:`Textstyle` for the given DXF entity."""
    if entity.doc is None:
        return Textstyle()
    try:
        name = entity.dxf.style
    except const.DXFAttributeError:
        name = "Standard"

    try:
        style = entity.doc.styles.get(name)
    except const.DXFTableEntryError:
        return Textstyle()
    return style



# Copyright (c) 2011-2019, Manfred Moitzi
# License: MIT License
import pytest
from io import StringIO
import ezdxf
from ezdxf.lldxf.tags import internal_tag_compiler
from ezdxf.lldxf.extendedtags import ExtendedTags
from ezdxf.sections.classes import ClassesSection
from ezdxf.lldxf.tagwriter import TagWriter
from ezdxf.tools.test import load_section
from ezdxf.entities import factory


@pytest.fixture(scope="module")
def section():
    sec = load_section(TESTCLASSES, "CLASSES")
    cls_entities = [factory.load(ExtendedTags(e)) for e in sec]
    return ClassesSection(None, iter(cls_entities))


def test_write(section):
    stream = StringIO()
    section.export_dxf(TagWriter(stream))
    result = stream.getvalue()
    stream.close()
    t1 = list(internal_tag_compiler(TESTCLASSES))
    t2 = list(internal_tag_compiler(result))
    assert t1 == t2


def test_empty_section():
    sec = load_section(EMPTYSEC, "CLASSES")
    cls_entities = [factory.load(ExtendedTags(e)) for e in sec]

    section = ClassesSection(None, iter(cls_entities))
    stream = StringIO()
    section.export_dxf(TagWriter(stream))
    result = stream.getvalue()
    stream.close()
    assert EMPTYSEC == result


def test_count_class_instances():
    def instance_count(name):
        return doc.classes.get(name).dxf.instance_count

    doc = ezdxf.new("R2004")

    doc.classes.add_class("IMAGE")
    doc.classes.add_class("IMAGEDEF")
    doc.classes.add_class("IMAGEDEF_REACTOR")
    doc.classes.add_class("RASTERVARIABLES")

    doc.classes.update_instance_counters()
    assert instance_count("IMAGE") == 0
    assert instance_count("IMAGEDEF") == 0
    assert instance_count("IMAGEDEF_REACTOR") == 0
    assert instance_count("RASTERVARIABLES") == 0

    image_def = doc.add_image_def("test", size_in_pixel=(400, 400))
    msp = doc.modelspace()
    msp.add_image(image_def, insert=(0, 0), size_in_units=(10, 10))

    doc.classes.update_instance_counters()
    assert instance_count("IMAGE") == 1
    assert instance_count("IMAGEDEF") == 1
    assert instance_count("IMAGEDEF_REACTOR") == 1
    assert instance_count("RASTERVARIABLES") == 1


EMPTYSEC = """  0
SECTION
  2
CLASSES
  0
ENDSEC
"""

TESTCLASSES = """  0
SECTION
  2
CLASSES
  0
CLASS
  1
ACDBDICTIONARYWDFLT
  2
AcDbDictionaryWithDefault
  3
ObjectDBX Classes
 90
        0
 91
        1
280
     0
281
     0
  0
CLASS
  1
DICTIONARYVAR
  2
AcDbDictionaryVar
  3
ObjectDBX Classes
 90
        0
 91
       13
280
     0
281
     0
  0
CLASS
  1
TABLESTYLE
  2
AcDbTableStyle
  3
ObjectDBX Classes
 90
     4095
 91
        1
280
     0
281
     0
0
ENDSEC
"""
